"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Feb 13, 2012.
"""
from everest.entities.utils import new_entity_id
from everest.querying.interfaces import IFilterSpecificationFactory
from everest.querying.specifications import FilterSpecificationFactory
from everest.repositories.memory import Aggregate
from everest.repositories.memory import Repository
from everest.testing import EntityTestCase
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.entities import MyEntityChild
from everest.tests.complete_app.entities import MyEntityParent
from mock import patch
from pyramid.threadlocal import get_current_registry
import gc
from everest.tests.complete_app.interfaces import IMyEntityParent
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.interfaces import IMyEntityChild
from everest.tests.complete_app.interfaces import IMyEntityGrandchild


__docformat__ = 'reStructuredText en'
__all__ = ['JoinedTransactionMemorySessionTestCase',
           'TransactionLessMemorySessionTestCase',
           ]


class _MemorySessionTestCaseBase(EntityTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'
    _session = None

    def set_up(self):
        EntityTestCase.set_up(self)
        # Some tests require the filter specification factory.
        flt_spec_fac = FilterSpecificationFactory()
        reg = get_current_registry()
        reg.registerUtility(flt_spec_fac, IFilterSpecificationFactory)
        self._repository = self._make_repository()
        # Strictly speaking, we should register the rcs only with *one*
        # repository; the tests do not require this at the moment, however.
        for ifc in (IMyEntityParent, IMyEntity, IMyEntityChild,
                    IMyEntityGrandchild):
            self._repository.register_resource(ifc)
        self._repository.initialize()
        self._session = self._repository.session_factory()

    def test_basics(self):
        ent = MyEntity()
        self.assert_is_none(ent.id)
        self.assert_is_none(ent.slug)
        # Load without ID fails.
        with self.assert_raises(ValueError) as cm:
            self._session.load(MyEntity, ent)
        self.assert_true(cm.exception.args[0].startswith('Can not load'))
        self._session.add(MyEntity, ent)
        self.assert_true(ent in self._session)
        self.assert_true(ent in self._session.query(MyEntity))
        self.assert_equal(list(self._session.new), [ent])
        # Commit triggers ID generation.
        self._session.commit()
        self.assert_is_not_none(ent.id)
        self.assert_is_not_none(ent.slug)
        # After commit, the session is empty.
        self.assert_false(ent in self._session)
        self.assert_equal(len(self._session.query(MyEntity).all()), 1)
        # Test loading by ID and slug.
        fetched_ent0 = self._session.get_by_id(MyEntity, ent.id)
        self.assert_equal(fetched_ent0.slug, ent.slug)
        fetched_ent1 = self._session.get_by_slug(MyEntity, ent.slug)
        self.assert_equal(fetched_ent1.id, ent.id)
        # We get a clone when we load an entity from the session.
        self.assert_false(fetched_ent0 is ent)
        # Once loaded, we always get the same entity.
        self.assert_true(fetched_ent0 is fetched_ent1)
        self._session.remove(MyEntity, fetched_ent0)
        self.assert_equal(len(self._session.query(MyEntity).all()), 0)
        self.assert_is_none(self._session.get_by_id(MyEntity, ent.id))
        self.assert_is_none(self._session.get_by_slug(MyEntity, ent.slug))
        self.assert_equal(list(self._session.deleted), [fetched_ent0])

    def test_remove_entity_not_in_session_raises_error(self):
        ent = MyEntity()
        self.assert_raises(ValueError, self._session.remove, MyEntity, ent)

    def test_add_deleted(self):
        ent = MyEntity()
        self._session.add(MyEntity, ent)
        self._session.commit()
        self._session.remove(MyEntity, ent)
        self._session.add(MyEntity, ent)

    def test_nested(self):
        ent = MyEntity()
        parent = MyEntityParent()
        ent.parent = parent
        child = MyEntityChild()
        ent.children.append(child)
        self._session.add(MyEntity, ent)
        self._session.commit()
        self.assert_equal(len(self._session.query(MyEntityChild).all()), 1)
        self.assert_equal(len(self._session.query(MyEntityParent).all()), 1)
        fetched_ent = self._session.query(MyEntity).one()
        self.assert_is_not_none(fetched_ent.parent)
        self.assert_equal(len(fetched_ent.children), 1)

    def test_nested_with_set_collection_type(self):
        ent = MyEntity()
        child = MyEntityChild()
        ent.children = set([child])
        self._session.add(MyEntity, ent)
        self._session.commit()
        fetched_ent = self._session.query(MyEntity).one()
        self.assert_true(isinstance(fetched_ent.children, set))

    def test_nested_with_invalid_collection_type(self):
        ent = MyEntity()
        child = MyEntityChild()
        ent.children = (child,)
        self.assert_raises(ValueError, self._session.add, MyEntity, ent)
        ent.id = 0
        child.id = 0
        with self.assert_raises(ValueError) as cm:
            self._session.load(MyEntity, ent)
        self.assert_true(cm.exception.args[0].startswith('Do not know'))

    def test_nested_with_invalid_collection_data(self):
        ent = MyEntity()
        ent.children = [None]
        self.assert_raises(ValueError, self._session.add, MyEntity, ent)

    def _make_repository(self):
        raise NotImplementedError('Abstract method.')


class JoinedTransactionMemorySessionTestCase(_MemorySessionTestCaseBase):
    def _make_repository(self):
        return Repository('DUMMY', Aggregate, join_transaction=True)


class TransactionLessMemorySessionTestCase(_MemorySessionTestCaseBase):
    def _make_repository(self):
        return Repository('DUMMY', Aggregate)

    def test_update_entity_not_in_session_raises_error(self):
        ent = MyEntity()
        self.assert_raises(ValueError, self._session.update, MyEntity, ent)

    def test_get_entity_not_in_session(self):
        self.assert_is_none(self._session.get_by_id(MyEntity, '-1'))

    def test_references(self):
        ent = MyEntity()
        self._session.add(MyEntity, ent)
        self.assert_equal(len(self._session.query(MyEntity).all()), 1)
        # Even with the last external ref gone, the cache should hold a
        # reference to the entities it manages.
        del ent
        gc.collect()
        self.assert_equal(len(self._session.query(MyEntity).all()), 1)

    def test_id_generation(self):
        ent1 = MyEntity()
        self._session.add(MyEntity, ent1)
        self._session.commit()
        self.assert_is_not_none(ent1.id)
        ent2 = MyEntity()
        self._session.add(MyEntity, ent2)
        self._session.commit()
        self.assert_is_not_none(ent2.id)
        # entity IDs can be sorted by creation time.
        self.assert_true(ent2.id > ent1.id)

    @patch('%s.entities.MyEntity.slug' %
           _MemorySessionTestCaseBase.package_name, None)
    def test_with_id_without_slug(self):
        ent = MyEntity(id=0)
        self._session.add(MyEntity, ent)
        self.assert_true(self._session.get_by_id(MyEntity, 0) is ent)

    @patch('%s.entities.MyEntity.slug' %
           _MemorySessionTestCaseBase.package_name, 'slug')
    def test_without_id_with_slug(self):
        ent = MyEntity()
        self._session.add(MyEntity, ent)
        self.assert_true(self._session.get_by_slug(MyEntity, 'slug') is ent)

    def test_duplicate_id_raises_error(self):
        ent_id = new_entity_id()
        ent1 = MyEntity(id=ent_id)
        self._session.add(MyEntity, ent1)
        ent2 = MyEntity(id=ent_id)
        self.assert_raises(ValueError, self._session.add, MyEntity, ent2)

    @patch('%s.entities.MyEntity.slug' %
           _MemorySessionTestCaseBase.package_name, 'slug')
    def test_duplicate_slug_raises_error(self):
        ent1 = MyEntity()
        self._session.add(MyEntity, ent1)
        ent2 = MyEntity()
        self.assert_raises(ValueError,
                           self._session.add, MyEntity, ent2)

    def test_cope_with_numeric_id(self):
        ent = MyEntity(id=0)
        self._session.add(MyEntity, ent)
        self.assert_equal(self._session.get_by_id(MyEntity, ent.id).id,
                          ent.id)
        self.assert_equal(self._session.get_by_slug(MyEntity, ent.slug).id,
                          ent.id)

    @patch('%s.entities.MyEntity.slug' %
           _MemorySessionTestCaseBase.package_name, 'slug')
    def test_repeated_add_remove(self):
        ent1 = MyEntity()
        self._session.add(MyEntity, ent1)
        self.assert_true(self._session.get_by_slug(MyEntity, ent1.slug)
                         is ent1)
        self._session.remove(MyEntity, ent1)
        self.assert_is_none(self._session.get_by_slug(MyEntity, ent1.slug))
        ent2 = MyEntity()
        self._session.add(MyEntity, ent2)
        self.assert_true(self._session.get_by_slug(MyEntity, ent2.slug)
                         is ent2)
        self._session.remove(MyEntity, ent2)
        self.assert_is_none(self._session.get_by_slug(MyEntity, ent2.slug))

    def test_remove_flush_add(self):
        ent = MyEntity()
        self._session.add(MyEntity, ent)
        self._session.commit()
        self.assert_equal(len(self._session.query(MyEntity).all()), 1)
        self._session.remove(MyEntity, ent)
        self.assert_equal(len(self._session.query(MyEntity).all()), 0)
        self._session.add(MyEntity, ent)
        self.assert_equal(len(self._session.query(MyEntity).all()), 1)

    def test_add_immediate_remove(self):
        ent1 = MyEntity()
        self._session.add(MyEntity, ent1)
        self._session.remove(MyEntity, ent1)
        self.assert_false(ent1 in self._session)
        self.assert_equal(len(self._session.query(MyEntity).all()), 0)

    @patch('%s.entities.MyEntity.slug' %
           _MemorySessionTestCaseBase.package_name, 'slug')
    def test_add_remove_add(self):
        ent1 = MyEntity()
        self._session.add(MyEntity, ent1)
        self.assert_true(ent1 in self._session)
        self._session.remove(MyEntity, ent1)
        self.assert_false(ent1 in self._session)
        self._session.add(MyEntity, ent1)
        self.assert_true(ent1 in self._session)

    def test_update_without_id_raises_error(self):
        ent1 = MyEntity(id=0)
        self._session.add(MyEntity, ent1)
        self._session.commit()
        # Re-load.
        ent2 = self._session.load(MyEntity, ent1)
        ent2.id = None
        with self.assert_raises(ValueError) as cm:
            self._session.commit()
        exc_msg = 'Could not persist data - target entity not found'
        self.assert_true(str(cm.exception).startswith(exc_msg))

    @patch('%s.entities.MyEntity.slug' %
           _MemorySessionTestCaseBase.package_name, 'slug')
    def test_update_with_different_slug(self):
        ent_id = 0
        ent1 = MyEntity(id=ent_id)
        self._session.add(MyEntity, ent1)
        ent2 = self._session.get_by_id(MyEntity, ent_id)
        text = 'foo'
        ent2.slug = text
        self._session.commit()
        ent3 = self._session.query(MyEntity).filter_by(slug=text).one()
        self.assert_equal(ent3.id, ent_id)

    def test_failing_commit_duplicate_id(self):
        ent1 = MyEntity()
        self._session.add(MyEntity, ent1)
        ent2 = MyEntity()
        self._session.add(MyEntity, ent2)
        self.assert_is_none(ent1.id)
        self.assert_is_none(ent2.id)
        ent2.id = ent1.id = 0
        self.assert_raises(ValueError, self._session.commit)

    @patch('%s.entities.MyEntity.slug' %
           _MemorySessionTestCaseBase.package_name, 'slug')
    def test_failing_flush_duplicate_slug(self):
        ent1 = MyEntity()
        self._session.add(MyEntity, ent1)
        ent2 = MyEntity()
        ent2.slug = None
        self._session.add(MyEntity, ent2)
        ent2.slug = 'slug'
        self.assert_raises(ValueError, self._session.commit)

    @patch('%s.entities.MyEntity.slug' %
           _MemorySessionTestCaseBase.package_name, 'slug')
    def test_find_added_by_id(self):
        ent1 = MyEntity(id=0)
        self._session.add(MyEntity, ent1)
        ent2 = self._session.get_by_id(MyEntity, ent1.id)
        self.assert_is_not_none(ent2)
        self.assert_equal(ent1.id, ent2.id)

    @patch('%s.entities.MyEntity.slug' %
           _MemorySessionTestCaseBase.package_name, 'slug')
    def test_find_added_by_slug(self):
        ent1 = MyEntity()
        self._session.add(MyEntity, ent1)
        ent2 = self._session.get_by_slug(MyEntity, ent1.slug)
        self.assert_is_not_none(ent2)
        self.assert_equal(ent1.id, ent2.id)

    @patch('%s.entities.MyEntity.slug' %
           _MemorySessionTestCaseBase.package_name, None)
    def test_find_added_with_none_slug_by_slug(self):
        ent1 = MyEntity()
        self._session.add(MyEntity, ent1)
        ent1.slug = 'testslug'
        ent2 = self._session.get_by_slug(MyEntity, ent1.slug)
        self.assert_is_not_none(ent2)
        self.assert_equal(ent1.id, ent2.id)

    def test_update(self):
        ent1 = MyEntity(id=0)
        self._session.add(MyEntity, ent1)
        ent2 = MyEntity()
        ent2.id = ent1.id
        my_attr_value = 1
        ent2.number = my_attr_value
        self._session.update(MyEntity, ent2)
        ent3 = self._session.get_by_id(MyEntity, ent1.id)
        self.assert_is_not_none(ent3)
        self.assert_equal(ent3.id, ent1.id)
        self.assert_equal(ent3.number, my_attr_value)



"""
Batch.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 15, 2011.
"""
import math

__docformat__ = "reStructuredText en"
__all__ = ['Batch',
           ]


class Batch(object):
    """
    Helper class to manage batches in a sequence.
    """
    def __init__(self, start, size, total_size):
        """
        :param int start: start index for this batch.
        :param int size: batch size.
        :param int total_size: total size of the batched sequence.
        """
        if start < 0:
            raise ValueError('Batch start must be zero or a positive number.')
        if not size > 0:
            raise ValueError('Batch size must be a positive number.')
        self.start = int(start) // int(size) * int(size)
        self.size = size
        self.total_size = total_size

    @property
    def next(self):
        """
        Returns the next batch for the batched sequence or `None`, if
        this batch is already the last batch.

        :rtype: :class:`Batch` instance or `None`.
        """
        if self.start + self.size > self.total_size:
            result = None
        else:
            result = Batch(self.start + self.size, self.size, self.total_size)
        return result

    @property
    def previous(self):
        """
        Returns the previous batch for the batched sequence or `None`, if
        this batch is already the first batch.

        :rtype: :class:`Batch` instance or `None`.
        """
        if self.start - self.size < 0:
            result = None
        else:
            result = Batch(self.start - self.size, self.size, self.total_size)
        return result

    @property
    def first(self):
        """
        Returns the first batch for the batched sequence.

        :rtype: :class:`Batch` instance.
        """
        return Batch(0, self.size, self.total_size)

    @property
    def last(self):
        """
        Returns the last batch for the batched sequence.

        :rtype: :class:`Batch` instance.
        """
        start = max(self.number - 1, 0) * self.size
        return Batch(start, self.size, self.total_size)

    @property
    def number(self):
        """
        Returns the number of batches the batched sequence contains.

        :rtype: integer.
        """
        return int(math.ceil(self.total_size / float(self.size)))

    @property
    def index(self):
        """
        Returns the index of this batch in the batched sequence.

        :rtype: integer
        """
        return int(self.start / self.size)



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Adapted from the ZCML unit tests in BFG.

Created on Jun 17, 2011.
"""
from everest.configuration import Configurator
from everest.entities.interfaces import IEntity
from everest.repositories.interfaces import IRepositoryManager
from everest.mime import CsvMime
from everest.representers.base import Representer
from everest.representers.interfaces import IRepresenter
from everest.representers.utils import as_representer
from everest.resources.base import Collection
from everest.resources.interfaces import ICollectionResource
from everest.resources.interfaces import IMemberResource
from everest.resources.interfaces import IService
from everest.testing import Pep8CompliantTestCase
from everest.tests import simple_app as package
from everest.tests.simple_app.entities import FooEntity
from everest.tests.simple_app.interfaces import IBar
from everest.tests.simple_app.interfaces import IFoo
from everest.tests.simple_app.resources import FooCollection
from everest.tests.simple_app.resources import FooMember
from pyramid.testing import setUp as set_up_testing
from pyramid.testing import tearDown as tear_down_testing
from pyramid.threadlocal import get_current_registry

__docformat__ = 'reStructuredText en'
__all__ = ['DirectivesTestCase',
           ]


class DirectivesTestCase(Pep8CompliantTestCase):

    def set_up(self):
        set_up_testing()
        reg = self._registry = get_current_registry()
        self._config = Configurator(registry=reg, package=package)
        self._config.setup_registry()
        repo_mgr = self._config.get_registered_utility(IRepositoryManager)
        repo_mgr.initialize_all()

    def tear_down(self):
        tear_down_testing()

    def test_configure_with_simple_zcml(self):
        # Load the configuration.
        self._config.load_zcml('everest.tests.simple_app:configure_simple.zcml')
        reg = self._registry
        # Check adapters.
        ent = FooEntity()
        member = object.__new__(FooMember)
        coll_cls = reg.queryUtility(IFoo, name='collection-class')
        self.assert_is_not_none(coll_cls)
        self.assert_is_not_none(coll_cls.root_name)
        self.assert_is_not_none(coll_cls.title)
        coll = object.__new__(coll_cls)
        self.__check(reg, member, ent, coll)
        # Check service.
        srvc = reg.queryUtility(IService)
        self.assert_is_not_none(srvc)
        srvc.start()
        self.assert_true(isinstance(srvc.get('foos'), Collection))
        self.assert_is_none(srvc.get(coll_cls))
        self.assert_is_none(srvc.get(IBar))

    def test_configure_with_full_zcml(self):
        reg = self._registry
        # Check adapters.
        ent = FooEntity()
        member = object.__new__(FooMember)
        coll = object.__new__(FooCollection)
        # Make sure no adapters are in the registry.
        self.assert_is_none(reg.queryAdapter(coll, IMemberResource,
                                          name='member-class'))
        self.assert_is_none(reg.queryAdapter(member, ICollectionResource,
                                          name='collection-class'))
        self.assert_is_none(reg.queryAdapter(member, IEntity,
                                          name='entity-class'))
        self.assert_is_none(reg.queryAdapter(ent, IMemberResource))
        self.assert_is_none(reg.queryAdapter(coll, IRepresenter,
                                          name=CsvMime.mime_type_string))
        # Load the configuration.
        config = Configurator(registry=reg, package=package)
        config.load_zcml('everest.tests.simple_app:configure.zcml')
        self.__check(reg, member, ent, coll)
        rpr = as_representer(coll, CsvMime)
        self.assert_true(isinstance(rpr, Representer))

    def test_configure_with_custom_repo_zcml(self):
        repo_mgr = self._registry.queryUtility(IRepositoryManager)
        self.assert_is_none(repo_mgr.get('CUSTOM_MEMORY'))
        self.assert_is_none(repo_mgr.get('CUSTOM_FILESYSTEM'))
        self.assert_is_none(repo_mgr.get('CUSTOM_RDB'))
        self._config.load_zcml('everest.tests.simple_app:configure_repos.zcml')
        self.assert_is_not_none(repo_mgr.get('CUSTOM_MEMORY'))
        self.assert_is_not_none(repo_mgr.get('CUSTOM_FILESYSTEM'))
        self.assert_is_not_none(repo_mgr.get('CUSTOM_RDB'))

    def __check(self, reg, member, ent, coll):
        for idx, obj in enumerate((member, coll, ent)):
            self.assert_equal(reg.queryAdapter(obj, IMemberResource,
                                              name='member-class'),
                              type(member))
            self.assert_equal(reg.queryAdapter(obj, ICollectionResource,
                                              name='collection-class'),
                              type(coll))
            self.assert_equal(reg.queryAdapter(obj, IEntity,
                                              name='entity-class'),
                              type(ent))
            if idx < 2: # lookup with class only for member/collection.
                self.assert_equal(reg.queryAdapter(type(obj), IMemberResource,
                                                  name='member-class'),
                                  type(member))
                self.assert_equal(reg.queryAdapter(type(obj),
                                                   ICollectionResource,
                                                  name='collection-class'),
                                  type(coll))
                self.assert_equal(reg.queryAdapter(type(obj), IEntity,
                                                   name='entity-class'),
                                  type(ent))
        # Check instance adapters.
        self.assert_is_not_none(reg.queryAdapter(ent, IMemberResource))



"""
Entity relationship.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 12, 2013.
"""
from everest.constants import CARDINALITY_CONSTANTS
from everest.constants import RELATION_OPERATIONS
from everest.relationship import RELATIONSHIP_DIRECTIONS
from everest.relationship import Relationship
from everest.utils import get_nested_attribute
from everest.utils import set_nested_attribute

__docformat__ = 'reStructuredText en'
__all__ = ['DomainRelationship',
           'LazyDomainRelationship',
           ]


class DomainRelationship(Relationship):
    """
    Relationship between entity domain objects.
    """
    def add(self, related, direction=None, safe=False):
        self.__action(related, RELATION_OPERATIONS.ADD, direction,
                      safe)

    def remove(self, related, direction=None, safe=False):
        self.__action(related, RELATION_OPERATIONS.REMOVE,
                      direction, safe)

    def _get_specification_attributes(self):
        return self.descriptor.entity_attr, self.descriptor.entity_backref

    def __action(self, related, rel_op, direction, safe):
        if direction is None:
            direction = self.direction
        crd = self.descriptor.cardinality
        if direction & RELATIONSHIP_DIRECTIONS.FORWARD \
           and not self.descriptor.entity_attr is None:
            self.__action_one_direction(crd.relatee, rel_op, self.relator,
                                        related, self.descriptor.entity_attr,
                                        safe)
        if direction & RELATIONSHIP_DIRECTIONS.REVERSE \
           and not self.descriptor.entity_backref is None:
            self.__action_one_direction(crd.relator, rel_op, related,
                                        self.relator,
                                        self.descriptor.entity_backref,
                                        safe)

    def __action_one_direction(self, cardinality_relatee, rel_op,
                               relator, related, attr_name, safe):
        if cardinality_relatee == CARDINALITY_CONSTANTS.ONE:
            if rel_op == RELATION_OPERATIONS.ADD:
                set_nested_attribute(relator, attr_name, related)
            elif rel_op == RELATION_OPERATIONS.REMOVE:
                set_nested_attribute(relator, attr_name, None)
        else:
            relatee = get_nested_attribute(relator, attr_name)
            if rel_op == RELATION_OPERATIONS.ADD:
                if not (safe and related in relatee):
                    # FIXME: Assuming a list here.
                    relatee.append(related)
            elif rel_op == RELATION_OPERATIONS.REMOVE:
                # FIXME: Assuming a list here.
                if safe:
                    try:
                        relatee.remove(related)
                    except ValueError:
                        pass
                else:
                    relatee.remove(related)


class LazyDomainRelationship(DomainRelationship):
    """
    Lazy version of a domain relationship.

    Operations are stored and only executed when the relationship is called.
    A lazy relationship can only be used repeatedly for the same operation
    with the same options.
    """
    __relator_proxy = None
    __action = None
    __kw = None

    #: This is the relatee set dynamically through calls to .add and .remove
    relatee = None

    def _set_relator(self, relator):
        self.__relator_proxy = relator

    def _get_relator(self):
        return self.__relator_proxy.get_entity()

    relator = property(_get_relator, _set_relator)

    def add(self, related, direction=None, safe=False):
        self.__lazy_action(DomainRelationship.add, related,
                           dict(direction=direction,
                                safe=safe))

    def remove(self, related, direction=None, safe=False):
        self.__lazy_action(DomainRelationship.remove, related,
                           dict(direction=direction,
                                safe=safe))

    def __call__(self):
        if self.descriptor.cardinality.relatee == CARDINALITY_CONSTANTS.ONE:
            self.__action(self, self.relatee, **self.__kw)
        else:
            for entity in self.relatee:
                self.__action(self, entity, **self.__kw)

    def __lazy_action(self, method, related, kw):
#        # FIXME: These checks ought to be enabled!
#        if not self.__action is None and method != self.__action:
#            raise ValueError('Must use the same action for repeated use '
#                             'of a lazy relationship.')
        self.__action = method
#        if not self.__kw is None and kw != self.__kw:
#            raise ValueError('Must use the same options for repeated use '
#                             'of a lazy relationship.')
        self.__kw = kw
        if self.descriptor.cardinality.relatee == CARDINALITY_CONSTANTS.ONE:
            self.relatee = related
        else:
            if self.relatee is None:
                self.relatee = []
            self.relatee.append(related)



"""
Querying functionality for the memory repository.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 7, 2013.
"""
from everest.querying.base import EXPRESSION_KINDS
from everest.querying.base import ExpressionBuilderMixin
from everest.querying.base import Query
from everest.querying.base import RepositoryQuery
from everest.querying.filtering import RepositoryFilterSpecificationVisitor
from everest.querying.interfaces import IFilterSpecificationVisitor
from everest.querying.interfaces import IOrderSpecificationVisitor
from everest.querying.ordering import RepositoryOrderSpecificationVisitor
from itertools import islice
from zope.interface import implementer # pylint: disable=E0611,F0401
import functools

__docformat__ = 'reStructuredText en'
__all__ = ['EvalFilterExpression',
           'EvalOrderExpression',
           'MemoryQuery',
           'MemoryRepositoryQuery',
           'ObjectFilterSpecificationVisitor',
           'ObjectOrderSpecificationVisitor',
           ]


class EvalFilterExpression(object):
    """
    Evaluation filter expression.
    """
    def __init__(self, spec):
        self.__spec = spec

    def __call__(self, entities):
        return self.__evaluator(self.__spec, entities)

    def __and__(self, other):
        return EvalFilterExpression(self.__spec & other.__spec) # pylint: disable=W0212

    def __or__(self, other):
        return EvalFilterExpression(self.__spec | other.__spec) # pylint: disable=W0212

    def __invert__(self):
        return EvalFilterExpression(~self.__spec)

    @staticmethod
    def __evaluator(spec, entities):
        return (ent for ent in entities if spec.is_satisfied_by(ent))


class EvalOrderExpression(object):
    """
    Evaluation order expression.
    """
    def __init__(self, spec):
        self.__spec = spec

    def __call__(self, entities):
        return sorted(entities, key=functools.cmp_to_key(self.__spec.cmp))

    def __and__(self, other):
        return EvalOrderExpression(self.__spec & other.__spec) # pylint: disable=W0212


class EvalExpressionBuilderMixin(ExpressionBuilderMixin):
    """
    Mixin class for building eval filter and order expressions from
    specifications.
    """
    expression_kind = EXPRESSION_KINDS.EVAL


class MemoryQuery(EvalExpressionBuilderMixin, Query):
    """
    Query operating on objects held in a sequence object.
    """
    def __init__(self, entity_class, entities):
        Query.__init__(self, entity_class)
        self.__entities = entities

    def __iter__(self):
        ents = iter(self.__entities)
        if not self._filter_expr is None:
            ents = self._filter_expr(ents)
        if not self._order_expr is None:
            # Ordering always involves a copy and conversion to a list, so
            # we have to wrap in an iterator.
            ents = iter(self._order_expr(ents))
        if not self._slice_key is None:
            ents = islice(ents, self._slice_key.start, self._slice_key.stop)
        return ents

    def count(self):
        ents = iter(self.__entities)
        if not self._filter_expr is None:
            ents = self._filter_expr(ents)
        return len(list(ents))


class MemoryRepositoryQuery(EvalExpressionBuilderMixin, RepositoryQuery):
    """
    Query operating on objects kept in a memory repository.
    """
    pass


@implementer(IFilterSpecificationVisitor)
class ObjectFilterSpecificationVisitor(RepositoryFilterSpecificationVisitor):
    """
    Filter specification visitor building an evaluator for in-memory
    filtering.
    """
    def _conjunction_op(self, spec, *expressions):
        return EvalFilterExpression(spec)

    def _disjunction_op(self, spec, *expressions):
        return EvalFilterExpression(spec)

    def _negation_op(self, spec, expression):
        return EvalFilterExpression(spec)

    def _starts_with_op(self, spec):
        return EvalFilterExpression(spec)

    def _ends_with_op(self, spec):
        return EvalFilterExpression(spec)

    def _contains_op(self, spec):
        return EvalFilterExpression(spec)

    def _contained_op(self, spec):
        return EvalFilterExpression(spec)

    def _equal_to_op(self, spec):
        return EvalFilterExpression(spec)

    def _less_than_op(self, spec):
        return EvalFilterExpression(spec)

    def _less_than_or_equal_to_op(self, spec):
        return EvalFilterExpression(spec)

    def _greater_than_op(self, spec):
        return EvalFilterExpression(spec)

    def _greater_than_or_equal_to_op(self, spec):
        return EvalFilterExpression(spec)

    def _in_range_op(self, spec):
        return EvalFilterExpression(spec)


@implementer(IOrderSpecificationVisitor)
class ObjectOrderSpecificationVisitor(RepositoryOrderSpecificationVisitor):
    """
    Order specification visitor building an evaluator for in-memory
    ordering.
    """
    def _conjunction_op(self, spec, *expressions):
        return lambda entities: sorted(entities,
                                       key=functools.cmp_to_key(spec.cmp))

    def _asc_op(self, spec):
        return lambda entities: sorted(entities,
                                       key=functools.cmp_to_key(spec.cmp))

    def _desc_op(self, spec):
        return lambda entities: sorted(entities,
                                       key=functools.cmp_to_key(spec.cmp))



"""
Unit tests for operators.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Sep 7, 2012.
"""
from everest.querying.operators import BinaryOperator
from everest.querying.operators import NullaryOperator
from everest.querying.operators import UnaryOperator
from everest.testing import Pep8CompliantTestCase

__docformat__ = 'reStructuredText en'
__all__ = ['OperatorTestCase',
           ]


class OperatorTestCase(Pep8CompliantTestCase):
    def test_arity(self):
        self.assert_equal(NullaryOperator.arity, 0)
        self.assert_equal(UnaryOperator.arity, 1)
        self.assert_equal(BinaryOperator.arity, 2)

    def test_derived(self):
        class MyNullaryOperator(NullaryOperator):
            @staticmethod
            def apply():
                return True
        self.assert_true(MyNullaryOperator.apply())



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 1, 2011.
"""
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.querying.specifications import FilterSpecificationFactory
from everest.repositories.rdb import SqlFilterSpecificationVisitor
from everest.repositories.rdb.testing import RdbTestCaseMixin
from everest.representers.config import IGNORE_OPTION
from everest.representers.config import RepresenterConfiguration
from everest.representers.config import WRITE_AS_LINK_OPTION
from everest.representers.mapping import SimpleMappingRegistry
from everest.resources.attributes import get_resource_class_attribute
from everest.resources.attributes import get_resource_class_attribute_iterator
from everest.resources.attributes import get_resource_class_attribute_names
from everest.resources.attributes import get_resource_class_attributes
from everest.resources.attributes import get_resource_class_collection_attribute_iterator
from everest.resources.attributes import get_resource_class_member_attribute_iterator
from everest.resources.attributes import get_resource_class_relationship_attribute_iterator
from everest.resources.attributes import get_resource_class_terminal_attribute_iterator
from everest.resources.attributes import is_resource_class_collection_attribute
from everest.resources.attributes import is_resource_class_member_attribute
from everest.resources.attributes import is_resource_class_resource_attribute
from everest.resources.attributes import is_resource_class_terminal_attribute
from everest.resources.base import Collection
from everest.resources.base import Member
from everest.resources.base import ResourceToEntityFilterSpecificationVisitor
from everest.resources.descriptors import attribute_alias
from everest.resources.descriptors import collection_attribute
from everest.resources.descriptors import member_attribute
from everest.resources.descriptors import terminal_attribute
from everest.resources.staging import create_staging_collection
from everest.resources.utils import get_member_class
from everest.resources.utils import get_root_collection
from everest.resources.utils import resource_to_url
from everest.testing import Pep8CompliantTestCase
from everest.testing import ResourceTestCase
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.entities import MyEntityChild
from everest.tests.complete_app.entities import MyEntityParent
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.interfaces import IMyEntityChild
from everest.tests.complete_app.interfaces import IMyEntityParent
from everest.tests.complete_app.resources import MyEntityMember
from everest.tests.complete_app.resources import MyEntityParentMember
from everest.tests.complete_app.testing import create_collection
from everest.tests.complete_app.testing import create_entity
from mock import patch
import datetime

__docformat__ = 'reStructuredText en'
__all__ = ['AttributesTestCase',
           'MemoryDescriptorsTestCase',
           'RdbDescriptorsTestCase'
           ]

ATTRIBUTE_NAMES = ['id', 'parent', 'children', 'text',
                   'text_rc', 'number', 'date_time', 'parent_text']


class AttributesTestCase(Pep8CompliantTestCase):
    def test_names(self):
        self.assert_equal(
                list(get_resource_class_attribute_names(MyEntityMember)),
                ATTRIBUTE_NAMES)

    def test_terminal_iterator(self):
        it = get_resource_class_terminal_attribute_iterator(MyEntityMember)
        self.assert_equal(set([attr.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL
                              for attr in it]),
                          set([True]))

    def test_resource_iterator(self):
        it = get_resource_class_relationship_attribute_iterator(MyEntityMember)
        self.assert_equal(set([attr.kind in
                               (RESOURCE_ATTRIBUTE_KINDS.MEMBER,
                                RESOURCE_ATTRIBUTE_KINDS.COLLECTION)
                              for attr in it]),
                          set([True]))

    def test_member_iterator(self):
        it = get_resource_class_member_attribute_iterator(MyEntityMember)
        self.assert_equal(set([attr.kind == RESOURCE_ATTRIBUTE_KINDS.MEMBER
                              for attr in it]),
                          set([True]))

    def test_collection_iterator(self):
        it = get_resource_class_collection_attribute_iterator(MyEntityMember)
        self.assert_equal(set([attr.kind == RESOURCE_ATTRIBUTE_KINDS.COLLECTION
                              for attr in it]),
                          set([True]))

    def test_types(self):
        attrs = iter(get_resource_class_attributes(MyEntityMember).values())
        attr0 = next(attrs)
        self.assert_equal(attr0.resource_attr, ATTRIBUTE_NAMES[0])
        self.assert_equal(attr0.kind, RESOURCE_ATTRIBUTE_KINDS.TERMINAL)
        self.assert_equal(attr0.entity_attr, 'id')
        self.assert_equal(attr0.attr_type, int)
        attr1 = next(attrs)
        self.assert_equal(attr1.resource_attr, ATTRIBUTE_NAMES[1])
        self.assert_equal(attr1.kind, RESOURCE_ATTRIBUTE_KINDS.MEMBER)
        self.assert_equal(attr1.entity_attr, 'parent')
        self.assert_equal(attr1.attr_type, IMyEntityParent)
        attr2 = next(attrs)
        self.assert_equal(attr2.resource_attr, ATTRIBUTE_NAMES[2])
        self.assert_equal(attr2.kind,
                          RESOURCE_ATTRIBUTE_KINDS.COLLECTION)
        self.assert_equal(attr2.entity_attr, 'children')
        self.assert_equal(attr2.attr_type, IMyEntityChild)
        attr3 = next(attrs)
        self.assert_equal(attr3.resource_attr, ATTRIBUTE_NAMES[3])
        self.assert_equal(attr3.kind, RESOURCE_ATTRIBUTE_KINDS.TERMINAL)
        self.assert_equal(attr3.entity_attr, 'text')
        self.assert_equal(attr3.attr_type, str)
        attr4 = next(attrs)
        self.assert_equal(attr4.resource_attr, ATTRIBUTE_NAMES[4])
        attr5 = next(attrs)
        self.assert_equal(attr5.resource_attr, ATTRIBUTE_NAMES[5])
        self.assert_equal(attr5.kind, RESOURCE_ATTRIBUTE_KINDS.TERMINAL)
        self.assert_equal(attr5.entity_attr, 'number')
        self.assert_equal(attr5.attr_type, int)
        self.assert_true(is_resource_class_member_attribute(MyEntityMember,
                                                            'parent'))
        self.assert_true(is_resource_class_collection_attribute(MyEntityMember,
                                                                'children'))
        self.assert_true(is_resource_class_resource_attribute(MyEntityMember,
                                                              'parent'))
        self.assert_true(is_resource_class_resource_attribute(MyEntityMember,
                                                              'children'))
        self.assert_true(isinstance(getattr(MyEntityMember, 'id'),
                                    terminal_attribute))

    def test_inheritance(self):
        class MyEntityDerivedMember(MyEntityMember):
            text = terminal_attribute(int, 'text')
        attr = get_resource_class_attribute(MyEntityDerivedMember, 'text')
        self.assert_equal(attr.kind,
                          RESOURCE_ATTRIBUTE_KINDS.TERMINAL)
        self.assert_equal(attr.entity_attr, 'text')
        self.assert_equal(attr.attr_type, int)

    def test_invalid_derived_descriptor(self):
        class my_descriptor(member_attribute):
            pass
        with self.assert_raises(TypeError) as cm:
            type('my_rc', (Member,),
                 dict(foo=my_descriptor(IMyEntityParent, 'foo')))
        exc_msg = 'Unknown resource attribute type'
        self.assert_true(str(cm.exception).startswith(exc_msg))

    def test_invalid_descriptor_parameters(self):
        self.assert_raises(ValueError,
                           terminal_attribute, 'not-a-type', 'foo')
        self.assert_raises(ValueError,
                           member_attribute, 'not-a-resource', 'foo')

    def test_entity_backref(self):
        attr = member_attribute(IMyEntityParent, entity_attr='foo')
        self.assert_is_none(attr.resource_backref)
        self.assert_is_none(attr.entity_backref)


class _DescriptorsTestCase(ResourceTestCase):
    package_name = 'everest.tests.complete_app'
    TEST_TEXT = 'TEST TEXT'
    UPDATED_TEXT = 'UPDATED TEXT'
    PARENT_MAPPING_OPTIONS = {WRITE_AS_LINK_OPTION:False,
                              IGNORE_OPTION:False}
    CHILDREN_MAPPING_OPTIONS = {WRITE_AS_LINK_OPTION:False,
                                IGNORE_OPTION:False}
    GRANDCHILDREN_MAPPING_OPTIONS = {WRITE_AS_LINK_OPTION:False,
                                     IGNORE_OPTION:False}

    def test_attribute_checkers(self):
        self.assert_true(is_resource_class_terminal_attribute(IMyEntity,
                                                              'text'))
        self.assert_true(is_resource_class_member_attribute(IMyEntity,
                                                            'parent'))
        self.assert_true(is_resource_class_collection_attribute(IMyEntity,
                                                                'children'))
        self.assert_true(is_resource_class_resource_attribute(IMyEntity,
                                                              'parent'))
        self.assert_true(is_resource_class_resource_attribute(IMyEntity,
                                                              'children'))
        attr_names = list(get_resource_class_attribute_names(MyEntityMember))
        self.assert_equal(attr_names, ATTRIBUTE_NAMES)
        it = get_resource_class_attribute_iterator(MyEntityMember)
        self.assert_equal([attr.resource_attr for attr in it],
                          ATTRIBUTE_NAMES)

    def test_terminal_access(self):
        entity = MyEntity()
        member = MyEntityMember.create_from_entity(entity)
        self.assert_true(isinstance(member.text, str))
        self.assert_true(isinstance(member.number, int))
        self.assert_true(isinstance(member.date_time, datetime.datetime))

    def test_member_access(self):
        parent = MyEntityParent()
        entity = MyEntity(parent=parent)
        member = MyEntityMember.create_from_entity(entity)
        self.assert_true(isinstance(member.parent, MyEntityParentMember))
        self.assert_true(member.parent.get_entity() is parent)

    def test_collection_access(self):
        parent = MyEntityParent(id=0)
        entity = MyEntity(id=0, parent=parent)
        coll = get_root_collection(IMyEntity)
        member = coll.create_member(entity)
        self.assert_true(isinstance(member.children, Collection))
        child_entity = MyEntityChild()
        member.children.create_member(child_entity)
        self.assert_equal(len(member.children), 1)

    def test_update_from_data_terminal(self):
        my_entity = create_entity()
        coll = create_staging_collection(IMyEntity)
        member = coll.create_member(my_entity)
        member.text = self.UPDATED_TEXT
        mp = self._make_mapping()
        data_el = mp.map_to_data_element(member)
        del member
        del my_entity
        my_entity = create_entity()
        coll = get_root_collection(IMyEntity)
        context = coll.create_member(my_entity)
        self.assert_equal(context.text, MyEntity.DEFAULT_TEXT)
        context.update(data_el)
        self.assert_equal(context.text, self.UPDATED_TEXT)

    def test_update_from_data_terminal_in_parent(self):
        my_entity = create_entity()
        my_entity.parent.text = self.UPDATED_TEXT
        coll = create_staging_collection(IMyEntity)
        member = coll.create_member(my_entity)
        mp = self._make_mapping()
        data_el = mp.map_to_data_element(member)
        del member
        del my_entity
        my_entity = create_entity()
        coll = get_root_collection(IMyEntity)
        context = coll.create_member(my_entity)
        self.assert_equal(context.parent.text, MyEntity.DEFAULT_TEXT)
        context.update(data_el)
        self.assert_equal(context.parent.text, self.UPDATED_TEXT)

    def test_update_from_data_terminal_in_child(self):
        my_entity = create_entity()
        my_entity.children[0].text = self.UPDATED_TEXT
        coll = create_staging_collection(IMyEntity)
        member = coll.create_member(my_entity)
        mp = self._make_mapping()
        data_el = mp.map_to_data_element(member)
        del member
        del my_entity
        my_entity = create_entity()
        coll = get_root_collection(IMyEntity)
        context = coll.create_member(my_entity)
        self.assert_equal(next(iter(context.children)).text,
                          MyEntity.DEFAULT_TEXT)
        context.update(data_el)
        self.assert_equal(next(iter(context.children)).text,
                          self.UPDATED_TEXT)

    def test_update_from_data_member(self):
        my_entity = create_entity()
        new_parent = MyEntityParent()
        new_parent.text = self.UPDATED_TEXT
        new_parent.id = 2
        my_entity.parent = new_parent
        coll = create_staging_collection(IMyEntity)
        member = coll.create_member(my_entity)
        mp = self._make_mapping()
        data_el = mp.map_to_data_element(member)
        del member
        del my_entity
        my_entity = create_entity()
        coll = get_root_collection(IMyEntity)
        context = coll.create_member(my_entity)
        self.assert_equal(context.parent.text, MyEntity.DEFAULT_TEXT)
        context.update(data_el)
        self.assert_equal(context.parent.text, self.UPDATED_TEXT)

    def test_update_from_data_member_from_link(self):
        my_entity = create_entity()
        new_parent = MyEntityParent()
        new_parent.text = self.UPDATED_TEXT
        new_parent.id = 2
        my_entity.parent = new_parent
        coll = create_staging_collection(IMyEntity)
        member = coll.create_member(my_entity)
        mp = self._make_mapping()
        attribute_options = {('parent',):{WRITE_AS_LINK_OPTION:True},
                             }
        mp_cloned = mp.clone(attribute_options=attribute_options)
        data_el = mp_cloned.map_to_data_element(member)
        # The linked-to parent needs to be in the root collection.
        my_entity.parent = None
        del member
        del my_entity
        parent_coll = get_root_collection(IMyEntityParent)
        parent_coll.create_member(new_parent)
        my_entity = create_entity()
        coll = get_root_collection(IMyEntity)
        context = coll.create_member(my_entity)
        self.assert_equal(context.parent.text, MyEntity.DEFAULT_TEXT)
        context.update(data_el)
        self.assert_equal(context.parent.text, self.UPDATED_TEXT)

    def test_update_from_data_delete_child(self):
        my_entity = create_entity()
        del my_entity.children[0]
        coll = create_staging_collection(IMyEntity)
        member = coll.create_member(my_entity)
        mp = self._make_mapping()
        data_el = mp.map_to_data_element(member)
        del member
        del my_entity
        my_entity = create_entity()
        coll = get_root_collection(IMyEntity)
        context = coll.create_member(my_entity)
        self.assert_equal(len(context.children), 1)
        context.update(data_el)
        self.assert_equal(len(context.children), 0)

    def test_update_from_data_delete_grandchild(self):
        my_entity = create_entity()
        del my_entity.children[0].children[0]
        coll = create_staging_collection(IMyEntity)
        member = coll.create_member(my_entity)
        mp = self._make_mapping()
        data_el = mp.map_to_data_element(member)
        del member
        del my_entity
        my_entity = create_entity()
        coll = get_root_collection(IMyEntity)
        context = coll.create_member(my_entity)
        self.assert_equal(len(next(iter(context.children)).children),
                          1)
        context.update(data_el)
        self.assert_equal(len(next(iter(context.children)).children),
                          0)

    def test_update_from_data_add_child(self):
        my_entity = create_entity()
        new_child = MyEntityChild()
        my_entity.children.append(new_child)
        if new_child.parent is None:
            new_child.parent = my_entity
        coll = create_staging_collection(IMyEntity)
        member = coll.create_member(my_entity)
        self.assert_equal(len(member.children), 2)
        mp = self._make_mapping()
        data_el = mp.map_to_data_element(member)
        del member
        del my_entity
#        import gc; gc.collect()
        my_entity = create_entity()
        coll = get_root_collection(IMyEntity)
        context = coll.create_member(my_entity)
        self.assert_equal(len(context.children), 1)
        context.update(data_el)
        self.assert_equal(len(context.children), 2)

    def test_add_with_data_element(self):
        my_entity = MyEntity()
        my_entity_parent = MyEntityParent()
        my_entity.parent = my_entity_parent
        coll = create_staging_collection(IMyEntity)
        member = coll.create_member(my_entity)
        mp = self._make_mapping()
        data_el = mp.map_to_data_element(member)
        del member
        root_coll = get_root_collection(IMyEntity)
        root_coll.add(data_el)
        self.assert_equal(len(coll), 1)

    def test_remove_with_data_element(self):
        my_entity = create_entity()
        coll = get_root_collection(IMyEntity)
        member = coll.create_member(my_entity)
        mp = self._make_mapping()
        data_el = mp.map_to_data_element(member)
        del member
        coll.remove(data_el)
        self.assert_equal(len(coll), 0)

    def test_nested_get(self):
        my_entity = create_entity()
        coll = create_staging_collection(IMyEntity)
        member = coll.create_member(my_entity)
        self.assert_equal(member.parent_text, MyEntityParent.DEFAULT_TEXT)

    def test_nested_set(self):
        ent = MyEntity()
        mb = MyEntityMember.create_from_entity(ent)
        self.assert_true(mb.parent is None)
        self.assert_raises(AttributeError, setattr, mb, 'parent_text', 'foo')

    def test_invalid_descriptors(self):
        with self.assert_raises(ValueError) as cm:
            collection_attribute(IMyEntity)
        exc_msg = 'may be None, but not both.'
        self.assert_true(str(cm.exception).endswith(exc_msg))

    def test_backref_only_collection(self):
        coll = create_collection()
        mb = next(iter(coll))
        child_mb = next(iter(mb.children))
        with patch('%s.resources.MyEntityChildMember.children.entity_attr'
                   % self.package_name, None):
            self.assert_equal(len(child_mb.children), 1)
            grandchild_mb = next(iter(child_mb.children))
            grandchild_mb.parent = None
            self.assert_equal(len(child_mb.children), 0)

    def test_basic_urls(self):
        my_entity = create_entity()
        coll = get_root_collection(IMyEntity)
        mb = coll.create_member(my_entity)
        exp_url = '/my-entities/0/'
        url = resource_to_url(mb)
        self.assert_true(url.endswith(exp_url))
        exp_url = '/my-entity-parents/0/'
        url = resource_to_url(mb.parent)
        self.assert_true(url.endswith(exp_url))
        exp_url = '/my-entity-children/?q=parent:equal-to:' \
                  '"http://0.0.0.0:6543/my-entities/0/"'
        url = resource_to_url(mb.children)
        self.assert_true(url.endswith(exp_url))
        mb_child = mb.children['0']
        self.assert_equal(mb_child.id, 0)
        exp_url = "/my-entity-grandchildren/?q=parent:equal-to:" \
                  "'http://0.0.0.0:6543/my-entity-children/0/'"
        url = resource_to_url(mb_child.children)

    @patch('%s.resources.MyEntityChildMember.children.resource_backref'
           % package_name, None)
    def test_no_backref_collection_url(self):
        my_entity = create_entity()
        coll = get_root_collection(IMyEntity)
        mb = coll.create_member(my_entity)
        mb_child = mb.children['0']
        exp_url = 'my-entity-grandchildren/?q=id:contained:0'
        url = resource_to_url(mb_child.children)
        self.assert_true(url.endswith(exp_url))

    def test_alias(self):
        ent = MyEntityParent()
        mb = MyEntityParentMember.create_from_entity(ent)
        alias_descr = getattr(MyEntityParentMember, 'text_alias')
        self.assert_true(isinstance(alias_descr, attribute_alias))
        self.assert_equal(mb.text_alias, mb.text)
        mb.text_alias = 'altered text'
        self.assert_equal(mb.text, mb.text_alias)

    def test_set_member(self):
        my_entity = create_entity()
        coll = get_root_collection(IMyEntity)
        mb = coll.create_member(my_entity)
        txt = 'FROBNIC'
        new_parent = MyEntityParent(text=txt)
        parent_coll = get_root_collection(IMyEntityParent)
        parent_mb = parent_coll.create_member(new_parent)
        self.assert_not_equal(mb.parent.text, txt)
        mb.parent = parent_mb
        self.assert_equal(mb.parent.text, txt)

    def _make_mapping(self):
        reg = SimpleMappingRegistry()
        mp_opts = {('parent',): self.PARENT_MAPPING_OPTIONS,
                   ('children',): self.CHILDREN_MAPPING_OPTIONS,
                   ('children', 'children'):
                                self.GRANDCHILDREN_MAPPING_OPTIONS,
                   }
        conf = RepresenterConfiguration(attribute_options=mp_opts)
        mp = reg.create_mapping(MyEntityMember, conf)
        reg.set_mapping(mp)
        return mp


class RdbDescriptorsTestCase(RdbTestCaseMixin, _DescriptorsTestCase):
    def test_filter_specification_visitor(self):
        coll = get_root_collection(IMyEntity)
        mb_cls = get_member_class(coll)
        my_entity = create_entity()
        member = coll.create_member(my_entity)
        spec_fac = FilterSpecificationFactory()
        specs = [
                # Terminal access.
                spec_fac.create_equal_to('text', self.TEST_TEXT),
                # Terminal access with different name in entity.
                spec_fac.create_equal_to('text_rc', self.TEST_TEXT),
                # Nested member access with different name in entity.
                spec_fac.create_equal_to('parent.text_rc', self.TEST_TEXT),
                # Nested collection access with different name in entity.
                spec_fac.create_equal_to('children.text_rc', self.TEST_TEXT),
                # Access with dotted entity name in rc attr declaration.
                spec_fac.create_equal_to('parent_text', self.TEST_TEXT),
                # Access to member.
                spec_fac.create_equal_to('parent',
                                         member.parent.get_entity()),
                ]
        expecteds = [('text', MyEntity.text.__eq__(self.TEST_TEXT)),
                     ('text_ent', MyEntity.text_ent.__eq__(self.TEST_TEXT)),
                     ('parent.text_ent',
                      MyEntity.parent.has(
                                    MyEntityParent.text_ent.__eq__(
                                                        self.TEST_TEXT))),
                     ('children.text_ent',
                      MyEntity.children.any(
                                    MyEntityChild.text_ent.__eq__(
                                                        self.TEST_TEXT))),
                     ('parent.text_ent',
                      MyEntity.parent.has(
                                    MyEntityParent.text_ent.__eq__(
                                                        self.TEST_TEXT))),
                     ('parent',
                      MyEntity.parent.__eq__(member.parent.get_entity())),
                     ]
        for spec, expected in zip(specs, expecteds):
            new_attr_name, expr = expected
            visitor = ResourceToEntityFilterSpecificationVisitor(mb_cls)
            spec.accept(visitor)
            new_spec = visitor.expression
            self.assert_equal(new_spec.attr_name, new_attr_name)
            visitor = SqlFilterSpecificationVisitor(MyEntity)
            new_spec.accept(visitor)
            self.assert_equal(str(visitor.expression), str(expr))
        invalid_spec = spec_fac.create_equal_to('foo', self.TEST_TEXT)
        vst = ResourceToEntityFilterSpecificationVisitor(mb_cls)
        self.assert_raises(AttributeError, invalid_spec.accept, vst)


class MemoryDescriptorsTestCase(_DescriptorsTestCase):
    config_file_name = 'configure_no_rdb.zcml'



"""
This file is part of the everest project. 
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 14, 2012.
"""
from everest.resources.base import Collection
from everest.resources.utils import get_collection_class
from everest.resources.utils import get_service
from everest.testing import ResourceTestCase
from everest.tests.complete_app.interfaces import IMyEntity
from zope.interface import Interface # pylint: disable=E0611,F0401

__docformat__ = 'reStructuredText en'
__all__ = ['ServiceTestCase',
           ]


class ServiceTestCase(ResourceTestCase):
    package_name = 'everest.tests.complete_app'
    config_file_name = 'configure_no_rdb.zcml'

    def set_up(self):
        ResourceTestCase.set_up(self)
        self.srv = get_service()

    def test_register_after_start_raises_error(self):
        self.assert_raises(RuntimeError, self.srv.register, IFooResource)

    def test__getitem__(self):
        self.assert_true(isinstance(self.srv['my-entities'],
                                    get_collection_class(IMyEntity)))

    def test_remove(self):
        num_colls = len(self.srv)
        self.srv.remove(IMyEntity)
        self.assert_raises(KeyError, self.srv.__getitem__, 'my-entities')
        self.assert_equal(len(self.srv), num_colls - 1)

    def test__iter__(self):
        for coll in self.srv:
            self.assert_true(isinstance(coll, Collection))

    def test_add_with_existing_name_raises_error(self):
        self.assert_raises(ValueError, self.srv.add, IMyEntity)


class IFooResource(Interface): # pylint: disable=W0232
    pass



"""
Data elements.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Apr 25, 2012
"""
from collections import OrderedDict

from pyramid.compat import iteritems_

from everest.constants import RESOURCE_KINDS
from everest.representers.converters import SimpleConverterRegistry
from everest.representers.interfaces import ICollectionDataElement
from everest.representers.interfaces import ILinkedDataElement
from everest.representers.interfaces import IMemberDataElement
from everest.representers.interfaces import IResourceDataElement
from everest.representers.utils import data_element_tree_to_string
from everest.resources.utils import provides_collection_resource
from everest.resources.utils import provides_member_resource
from everest.resources.utils import resource_to_url
from zope.interface import implementer # pylint: disable=E0611,F0401
from zope.interface import providedBy as provided_by # pylint: disable=E0611,F0401
from everest.constants import RESOURCE_ATTRIBUTE_KINDS


__docformat__ = 'reStructuredText en'
__all__ = ['CollectionDataElement',
           'DataElement',
           'DataElementAttributeProxy',
           'LinkedDataElement',
           'MemberDataElement',
           'SimpleCollectionDataElement',
           'SimpleLinkedDataElement',
           'SimpleMemberDataElement',
           ]


@implementer(IResourceDataElement)
class DataElement(object):
    """
    Abstract base class for data element classes.

    Data elements manage value state during serialization and deserialization.
    Implementations may need to be adapted to the format of the external
    representation they manage.
    """
    #: Static attribute mapping.
    mapping = None

    @classmethod
    def create(cls):
        """
        Basic factory method.
        """
        inst = cls()
        return inst

    @classmethod
    def create_from_resource(cls, resource):
        """
        (Abstract) factory method taking a resource as input.
        """
        raise NotImplementedError('Abstract method.')

    def __str__(self):
        return data_element_tree_to_string(self)


@implementer(IMemberDataElement)
class MemberDataElement(DataElement):
    """
    Abstract base class for member data element classes.
    """
    #: Registry of representation string <-> value converters. To be set
    #: in derived classes.
    converter_registry = None

    def iterator(self):
        """
        Returns an iterator yielding name value pairs for every attribute set
        on this data element.

        :returns: Sequence of `attribute repr name, attribute value` 2-tuples.
        """
        raise NotImplementedError('Abstract method.')

    @property
    def data(self):
        """
        Returns an ordered dictionary constructed from the return values
        of the :method:`iterator` method.

        :returns: Ordered dictionary mapping attribute repr names to attribute
            values.
        """
        raise NotImplementedError('Abstract method.')

    def get_attribute(self, attr_name):
        """
        Returns the value for the given attribute representation name.

        :returns: Attribute value (of the type specified by the resource
            attribute).
        :raises AttributeError: If the attribute was not set on this data
            element.
        """
        raise NotImplementedError('Abstract method.')

    def set_attribute(self, attr_name, value):
        """
        Sets the value for the given attribute representation name to the
        given value.

        :raises AttributeError: if the underlying mapping does not
            have an attribute with the given representation name.
        """
        raise NotImplementedError('Abstract method.')

    def get_terminal(self, attr):
        """
        Returns the value for the given mapped terminal resource attribute.

        :param attr: attribute to retrieve.
        :type attr: :class:`everest.representers.attributes.MappedAttribute`
        :returns: Attribute value (of the type specified by the resource
            attribute) or `None` if no value is found for the given attribute
            name.
        """
        raise NotImplementedError('Abstract method.')

    def set_terminal(self, attr, value):
        """
        Sets the value for the given mapped terminal resource attribute.

        :type attr: :class:`everest.representers.attributes.MappedAttribute`
        :param value: Value of the attribute to set.
        """
        raise NotImplementedError('Abstract method.')

    def get_nested(self, attr):
        """
        Returns the mapped nested resource attribute (either a member or a
        collection resource attribute).

        :type attr: :class:`everest.representers.attributes.MappedAttribute`
        :returns: Object implementing `:class:IDataelement` or
          `None` if no nested resource is found for the given attribute name.
        """
        raise NotImplementedError('Abstract method.')

    def set_nested(self, attr, data_element):
        """
        Sets the value for the given mapped nested resource attribute (either
        a member or a collection resource attribute).

        :type attr: :class:`everest.representers.attributes.MappedAttribute`
        :param data_element: :class:DataElement or :class:LinkedDataElement
          object containing nested resource data.
        """
        raise NotImplementedError('Abstract method.')


@implementer(ICollectionDataElement)
class CollectionDataElement(DataElement):
    """
    Abstract base class for collection data elements.
    """
    def add_member(self, data_element):
        """
        Adds the given member data element to this collection data element.
        """
        raise NotImplementedError('Abstract method.')

    def get_members(self):
        """
        Returns all member data elements added to this collection data element.
        """
        raise NotImplementedError('Abstract method.')

    def __len__(self):
        """
        Returns the number of member data elements in this collection data
        element.
        """
        raise NotImplementedError('Abstract method.')


class _SimpleDataElementMixin(object):
    @classmethod
    def create_from_resource(cls, resource): # ignore resource pylint:disable=W0613,W0221
        return cls()


class SimpleMemberDataElement(_SimpleDataElementMixin, MemberDataElement):
    """
    Basic implementation of a member data element.
    """
    converter_registry = SimpleConverterRegistry

    __data = None

    def iterator(self):
        return iter(iteritems_(self.data))

    @property
    def data(self):
        if self.__data is None:
            self.__data = OrderedDict()
        return self.__data

    def get_attribute(self, attr_name):
        try:
            value = self.__data[attr_name]
        except KeyError:
            raise AttributeError(attr_name)
        return value

    def set_attribute(self, attr_name, value):
        attr = self.mapping.get_attribute_by_repr(attr_name)
        if attr.kind != RESOURCE_ATTRIBUTE_KINDS.TERMINAL \
           and not (isinstance(value, DataElement) or value is None):
            raise ValueError('Need a data element or None as attribute '
                             'value.')
        self.__data[attr_name] = value

    @property
    def terminals(self):
        if self.__data is None:
            self.__data = OrderedDict()
        return OrderedDict((k, v)
                           for (k, v) in iteritems_(self.__data)
                           if not isinstance(v, DataElement))

    @property
    def nesteds(self):
        if self.__data is None:
            self.__data = OrderedDict()
        return OrderedDict((k, v)
                           for (k, v) in iteritems_(self.__data)
                           if isinstance(v, DataElement))

    def get_nested(self, attr):
        return self.data.get(attr.repr_name)

    def set_nested(self, attr, data_element):
        self.data[attr.repr_name] = data_element

    def get_terminal(self, attr):
        return self.data.get(attr.repr_name)

    def set_terminal(self, attr, value):
        self.data[attr.repr_name] = value

    def get_terminal_converted(self, attr):
        """
        Returns the value of the specified attribute converted to a
        representation value.

        :param attr: Attribute to retrieve.
        :type attr: :class:`everest.representers.attributes.MappedAttribute`
        :returns: Representation string.
        """
        value = self.data.get(attr.repr_name)
        return self.converter_registry.convert_to_representation(
                                                            value,
                                                            attr.value_type)

    def set_terminal_converted(self, attr, repr_value):
        """
        Converts the given representation value and sets the specified
        attribute value to the converted value.

        :param attr: Attribute to set.
        :param str repr_value: String value of the attribute to set.
        """
        value = self.converter_registry.convert_from_representation(
                                                            repr_value,
                                                            attr.value_type)
        self.data[attr.repr_name] = value


class SimpleCollectionDataElement(_SimpleDataElementMixin,
                                  CollectionDataElement):
    """
    Basic implementation of a collection data element.
    """
    __members = None

    def add_member(self, data_element):
        self.members.append(data_element)

    def get_members(self):
        return self.members

    @property
    def members(self):
        if self.__members is None:
            self.__members = []
        return self.__members

    def __len__(self):
        return len(self.__members)


@implementer(ILinkedDataElement)
class LinkedDataElement(DataElement):
    """
    Data element managing a linked resource during serialization and
    deserialization.
    """
    @classmethod
    def create(cls, url, kind,
               id=None, relation=None, title=None, **options): # pylint: disable=W0622
        raise NotImplementedError('Abstract method.')

    @classmethod
    def create_from_resource(cls, resource):
        raise NotImplementedError('Abstract method.')

    def get_url(self):
        raise NotImplementedError('Abstract method.')

    def get_kind(self):
        raise NotImplementedError('Abstract method.')

    def get_id(self):
        raise NotImplementedError('Abstract method.')

    def get_relation(self):
        raise NotImplementedError('Abstract method.')

    def get_title(self):
        raise NotImplementedError('Abstract method.')


class SimpleLinkedDataElement(LinkedDataElement):
    """
    Basic implementation of a linked data element.
    """
    __url = None
    __kind = None
    __id = None
    __relation = None
    __title = None

    @classmethod
    def create(cls, url, kind,
               id=None, relation=None, title=None, **options): # pylint: disable=W0622
        inst = cls()
        # pylint: disable=W0212
        inst.__url = url
        inst.__kind = kind
        inst.__id = id
        inst.__relation = relation
        inst.__title = title
        # pylint: enable=W0212
        return inst

    @classmethod
    def create_from_resource(cls, resource):
        if provides_member_resource(resource):
            kind = RESOURCE_KINDS.MEMBER
            opts = dict(id=resource.id)
        elif provides_collection_resource(resource):
            kind = RESOURCE_KINDS.COLLECTION
            opts = {}
        else:
            raise ValueError('"%s" is not a resource.' % resource)
        return cls.create(resource_to_url(resource), kind,
                          relation=resource.relation,
                          title=resource.title,
                          **opts)

    def get_url(self):
        return self.__url

    def get_kind(self):
        return self.__kind

    def get_id(self):
        return self.__id

    def get_relation(self):
        return self.__relation

    def get_title(self):
        return self.__title


class DataElementAttributeProxy(object):
    """
    Convenience proxy for accessing data from data elements.

    The proxy allows you to transparently access terminal, member, and
    collection attributes. Nested access is also supported.

    Example: ::

       prx = DataElementAttributeProxy(data_element)
       de_id = prx.id                              # terminal access
       de_parent = prx.parent                      # member access
       de_child = prx.children[0]                  # collection access
       de_grandchild = prx.children[0].children[0] # nested collection access
    """
    def __init__(self, data_element):
        if ILinkedDataElement in provided_by(data_element):
            raise ValueError('Do not use data element proxies with linked '
                             'data elements.')
        self.__data_element = data_element

    def get_data_element(self):
        """
        Returns the wrapped data element. Useful for proxies returned from
        accessing nested attributes.
        """
        return self.__data_element

    def __getattr__(self, name):
        value = self.__data_element.get_attribute(name)
        ifcs = provided_by(value)
        if IMemberDataElement in ifcs:
            value = DataElementAttributeProxy(value)
        elif ICollectionDataElement in ifcs:
            value = [DataElementAttributeProxy(mb_el)
                     for mb_el in value.get_members()]
        return value

    def __setattr__(self, name, value):
        # Avoid recursion for setting instance data during constructor call.
        if name in ['_DataElementAttributeProxy__data_element',
                    '_DataElementAttributeProxy__data',
                    '_DataElementAttributeProxy__attr_map']:
            self.__dict__[name] = value
        else:
            self.__data_element.set_attribute(name, value)
#            try:
#                attr = self.__attr_map[name]
#            except KeyError:
#                raise AttributeError(name)
#            else:
#                if attr.kind == RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
#                    self.__data_element.set_terminal(attr, value)
#                else:
#                    if not (isinstance(value, DataElement) or value is None):
#                        raise ValueError('Need a data element or None as '
#                                         'attribute value.')
#                    self.__data_element.set_nested(attr, value)



"""
Aggregate for the in-memory and filesystem backends.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 7, 2013.
"""
from everest.entities.base import RootAggregate
from everest.querying.base import EXPRESSION_KINDS

__docformat__ = 'reStructuredText en'
__all__ = ['MemoryAggregate',
           ]


class MemoryAggregate(RootAggregate):
    """
    Root aggregate implementation for the in-memory repository.

    :note: When entities without a slug are added to a memory aggregate, they
           can not be retrieved using the :meth:`get_by_slug` method.
    """
    _expression_kind = EXPRESSION_KINDS.EVAL



"""
In-memory repository.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jan 7, 2013.
"""
from everest.entities.utils import get_entity_class
from everest.entities.utils import new_entity_id
from everest.repositories.base import Repository
from everest.repositories.memory.aggregate import MemoryAggregate
from everest.repositories.memory.cache import EntityCacheMap
from everest.repositories.memory.session import MemorySessionFactory
from everest.repositories.state import ENTITY_STATUS
from threading import RLock

__docformat__ = 'reStructuredText en'
__all__ = ['MemoryRepository',
           ]


class MemoryRepository(Repository):
    """
    A repository that caches entities in memory.
    """
    _configurables = Repository._configurables \
                     + ['cache_loader']

    lock = RLock()

    def __init__(self, name, aggregate_class=None,
                 join_transaction=False, autocommit=False):
        if aggregate_class is None:
            aggregate_class = MemoryAggregate
        Repository.__init__(self, name, aggregate_class,
                            join_transaction=join_transaction,
                            autocommit=autocommit)
        self.__cache_map = EntityCacheMap()
        # By default, we do not use a cache loader.
        self.configure(cache_loader=None)

    def retrieve(self, entity_class, filter_expression=None,
                 order_expression=None, slice_key=None):
        cache = self.__get_cache(entity_class)
        return cache.retrieve(filter_expression=filter_expression,
                              order_expression=order_expression,
                              slice_key=slice_key)

    def flush(self, unit_of_work):
        for state in unit_of_work.iterator():
            if state.is_persisted:
                continue
            else:
                self.__persist(state)
                unit_of_work.mark_persisted(state.entity)

    def commit(self, unit_of_work):
        self.flush(unit_of_work)

    def rollback(self, unit_of_work):
        for state in unit_of_work.iterator():
            if state.is_persisted:
                self.__rollback(state)

    def __persist(self, state):
        source_entity = state.entity
        cache = self.__get_cache(type(source_entity))
        status = state.status
        if status == ENTITY_STATUS.NEW:
            # Autogenerate new ID.
            if source_entity.id is None:
                source_entity.id = new_entity_id()
            cache.add(source_entity)
        else:
            target_entity = cache.get_by_id(source_entity.id)
            if target_entity is None:
                raise ValueError('Could not persist data - target entity not '
                                 'found (ID used for lookup: %s).'
                                 % source_entity.id)
            if status == ENTITY_STATUS.DELETED:
                cache.remove(target_entity)
            elif status == ENTITY_STATUS.DIRTY:
                cache.update(state.data, target_entity)

    def __rollback(self, state):
        source_entity = state.entity
        cache = self.__get_cache(type(source_entity))
        if state.status == ENTITY_STATUS.DELETED:
            cache.add(source_entity)
        else:
            if state.status == ENTITY_STATUS.NEW:
                cache.remove(source_entity)
            elif state.status == ENTITY_STATUS.DIRTY:
                target_entity = cache.get_by_id(source_entity.id)
                cache.update(state.clean_data, target_entity)

    def _initialize(self):
        pass

    def _reset(self):
        self.__cache_map.clear()

    def _make_session_factory(self):
        return MemorySessionFactory(self)

    def __get_cache(self, entity_class):
        run_loader = not entity_class in self.__cache_map
        if run_loader:
            is_top_level = len(self.__cache_map.keys()) == 0
            self.__load_entities(entity_class, is_top_level)
        return self.__cache_map[entity_class]

    def __load_entities(self, entity_class, is_top_level):
        # Check if we have an entity loader configured.
        loader = self.configuration['cache_loader']
        if not loader is None:
            cache = self.__cache_map[entity_class]
            for ent in loader(entity_class):
                if ent.id is None:
                    ent.id = new_entity_id()
                cache.add(ent)
            # To fully initialize the cache, we also need to load collections
            # that are not linked to from any of the entities just loaded.
            if is_top_level:
                for reg_rc in self.registered_resources:
                    reg_ent_cls = get_entity_class(reg_rc)
                    if not reg_ent_cls in self.__cache_map:
                        self.__load_entities(reg_ent_cls, False)



"""
Entity and aggregate base classes.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 12, 2011.
"""
from everest.constants import RELATIONSHIP_DIRECTIONS
from everest.constants import RELATION_OPERATIONS
from everest.entities.interfaces import IAggregate
from everest.entities.interfaces import IEntity
from everest.exceptions import NoResultsException
from everest.querying.utils import get_filter_specification_factory
from everest.utils import get_filter_specification_visitor
from everest.utils import get_order_specification_visitor
from zope.interface import implementer # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['Aggregate',
           'Entity',
           ]


@implementer(IEntity)
class Entity(object):
    """
    Abstract base class for all model entities.

    All entities have an ID which is used as the default value for equality
    comparison. The object may be initialized without an ID.
    """
    #: The (unique) entity ID (integer or string).
    id = None

    def __init__(self, id=None): # redefining id pylint: disable=W0622
        if self.__class__ is Entity:
            raise NotImplementedError('Abstract class.')
        if not id is None:
            self.id = id

    @property
    def slug(self):
        """
        Returns a human-readable and URL-compatible string that is unique
        among all siblings of this entity.
        """
        return None if self.id is None else str(self.id)

    @classmethod
    def create_from_data(cls, data):
        """
        Creates a new instance of this entity from the given data map.
        """
        return cls(**data)

    def __eq__(self, other):
        return id(self) == id(other) \
               or isinstance(other, self.__class__) \
               and self.id == other.id \
               and not (self.id is None and other.id is None)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return id(self)


@implementer(IAggregate)
class Aggregate(object):
    """
    Abstract base class for aggregates.

    An aggregate is an accessor for a set of entities of the same type.

    Supports filtering, sorting, slicing, counting, iteration as well as
    retrieving, adding and removing entities.
    """
    #: Entity class (type) of the entities in this aggregate.
    entity_class = None

    def __init__(self):
        if self.__class__ is Aggregate:
            raise NotImplementedError('Abstract class.')
        #: Specification for filtering
        #: (:class:`everest.querying.specifications.FilterSpecification`).
        #: Attribute names in this specification are relative to the entity.
        self._filter_spec = None
        #: Specification for ordering
        #: (:class:`everest.querying.specifications.OrderSpecification`).
        #: Attribute names in this specification are relative to the entity.
        self._order_spec = None
        #: Key for slicing. (:type:`slice`).
        self._slice_key = None

    def clone(self):
        """
        Returns a clone of this aggregate.
        """
        clone = self.__class__.__new__(self.__class__)
        # Access protected member pylint: disable=W0212
        clone._filter_spec = self._filter_spec
        clone._order_spec = self._order_spec
        clone._slice_key = self._slice_key
        # pylint: enable=W0212
        return clone

    def iterator(self):
        """
        Returns an iterator for the entities contained in the underlying
        aggregate.

        If specified, filter, order, and slice settings are applied.

        :returns: An iterator for the aggregate entities.
        """
        return iter(self._get_ordered_query(None))

    def __iter__(self):
        return self.iterator()

    def count(self):
        """
        Returns the total number of entities in the underlying aggregate.
        If specified, filter specs are applied. A specified slice key is
        ignored.

        :returns: Number of aggregate members (:class:`int`).
        """
        return self._get_filtered_query(None).count()

    def get_by_id(self, id_key):
        """
        Returns an entity by ID  or `None` if the entity is not found.

        :note: If a filter is set which matches the requested entity, it
          will not be found.
        :param id_key: ID value to look up.
        :type id_key: `int` or `str`
        :raises: :class:`everest.exceptions.MultipleResultsException` if more
          than one entity is found for the given ID value.
        :returns: Specified entity or `None`.
        """
        raise NotImplementedError('Abstract method.')

    def get_by_slug(self, slug):
        """
        Returns an entity by slug or `None` if the entity is not found.

        :note: If a filter is set which matches the requested entity, it
          will not be found.
        :param slug: Slug value to look up.
        :type slug: `str`
        :raises: :class:`everest.exceptions.MultipleResultsException` if more
          than one entity is found for the given ID value.
        :returns: Entity or `None`
        """
        raise NotImplementedError('Abstract method.')

    def add(self, data):
        """
        Adds the given entity data to the aggregate.

        If the entity has an ID, it must be unique within the aggregate.

        :param data: Any object that can be adapted to
          :class:`everest.interfaces.IDataTraversalProxyAdapter`.
        :type entity: Object implementing
          :class:`everest.entities.interfaces.IEntity`.
        :raise ValueError: If an entity with the same ID exists.
        """
        raise NotImplementedError('Abstract method.')

    def remove(self, data):
        """
        Removes the given entity data from the aggregate.

        :param data: Any object that can be adapted to
          :class:`everest.interfaces.IDataTraversalProxyAdapter`.
        :raise ValueError: If the given entity was not found.
        """
        raise NotImplementedError('Abstract method.')

    def update(self, data, target=None):
        """
        Updates an existing entity data with the given data.

        Relies on the underlying repository for the implementation of the
        state update.

        :param data: Source entity data for the update.
        :type data: Any object that can be adapted to
          :class:`everest.interfaces.IDataTraversalProxyAdapter`.
        :param target: Target entity to transfer state to. If this is not
          given, the target is looked up by the ID of the given source
          data.
        :type source_entity: Object implementing
          :class:`everest.entities.interfaces.IEntity`.
        """
        raise NotImplementedError('Abstract method.')

    def query(self, **options):
        """
        Returns a query for this aggregate.
        """
        raise NotImplementedError('Abstract method.')

    def sync_with_repository(self):
        """
        Flushes all pending state to the repository and updates all loaded
        entities.
        """
        raise NotImplementedError('Abstract method.')

    @property
    def expression_kind(self):
        """
        Returns the kind of filter and order expression this aggregate builds.
        """
        raise NotImplementedError('Abstract property.')

    def get_root_aggregate(self, rc):
        """
        Returns a root aggregate for the given registered resource.

        The aggregate is retrieved from the same repository that was used to
        create this aggregate.
        """
        raise NotImplementedError('Abstract method.')

    def _get_filter(self):
        #: Returns the filter specification for this aggregate.
        return self._filter_spec

    def _set_filter(self, filter_spec):
        #: Sets the filter specification for this aggregate.
        self._filter_spec = filter_spec

    filter = property(_get_filter, _set_filter)

    def _get_order(self):
        #: Returns the order specification for this aggregate.
        return self._order_spec

    def _set_order(self, order_spec):
        #: Sets the order specification for this aggregate.
        self._order_spec = order_spec

    order = property(_get_order, _set_order)

    def _get_slice(self):
        #: Returns the slice key for this aggregate.
        return self._slice_key

    def _set_slice(self, slice_key):
        #: Sets the slice key for this aggregate. Filter and order specs
        #: are applied before the slicing operation is performed.
        self._slice_key = slice_key

    slice = property(_get_slice, _set_slice)

    def _query_optimizer(self, query, slice_key): # unused pylint: disable=W0613
        """
        Override this to generate optimized queries based on the given
        slice key.

        This default implementation just returns the given query as is.

        :param query: Query to optimize as returned from the data source.
        :param key: Slice key to use for the query or `None`, if no slicing
          was applied.
        """
        return query

    def _filter_visitor_factory(self):
        """
        Override this to create filter visitors with custom clauses.
        """
        visitor_cls = get_filter_specification_visitor(self.expression_kind)
        return visitor_cls(self.entity_class)

    def _order_visitor_factory(self):
        """
        Override this to create order visitors with custom clauses.
        """
        visitor_cls = get_order_specification_visitor(self.expression_kind)
        return visitor_cls(self.entity_class)

    def _get_filtered_query(self, key):
        #: Returns a query filtered by the current filter specification.
        query = self._query_optimizer(self.query(), key)
        query = self.__filter_query(query)
        return self.__slice_query(query)

    def _get_ordered_query(self, key):
        #: Returns a filtered query ordered by the current order
        #: specification.
        query = self._query_optimizer(self.query(), key)
        query = self.__filter_query(query)
        query = self.__order_query(query)
        return self.__slice_query(query)

    def __filter_query(self, query):
        if not self.filter is None:
            vst = self._filter_visitor_factory()
            self.filter.accept(vst)
            query = vst.filter_query(query)
        return query

    def __order_query(self, query):
        if not self._order_spec is None:
            vst = self._order_visitor_factory()
            self._order_spec.accept(vst)
            query = vst.order_query(query)
        return query

    def __slice_query(self, query):
        if not self._slice_key is None:
            query = query.slice(self._slice_key.start,
                                self._slice_key.stop)
        return query


class RootAggregate(Aggregate):
    """
    Abstract base class for root aggregates.

    A root aggregate provides access to all entities of a particular type
    in an underlying repository. It also holds a session factory which creates
    a thread-local session that stages operations on the repository.
    """
    #: This holds the value for the expression_kind property (the kind of
    #: filter and order expression this root aggregate builds).
    _expression_kind = None

    def __init__(self, entity_class, session_factory, repository):
        """
        Constructor.

        :param entity_class: The entity class (type) of the entities in this
            aggregate.
        :type entity_class: A class implementing
            :class:`everest.entities.interfaces.IEntity`.
        :param session_factory: The session factory for this aggregate.
        :param repository: The repository that created this aggregate.
        """
        if self.__class__ is RootAggregate:
            raise NotImplementedError('Abstract class.')
        Aggregate.__init__(self)
        #: The entity class managed by this aggregate.
        self.entity_class = entity_class
        #: The session factory.
        self._session_factory = session_factory
        # The repository that holds the entities.
        self.__repository = repository

    @classmethod
    def create(cls, entity_class, session_factory, repository):
        """
        Factory class method to create a new aggregate.
        """
        return cls(entity_class, session_factory, repository)

    def clone(self):
        clone = Aggregate.clone(self)
        clone.entity_class = self.entity_class
         # protected pylint: disable=W0212
        clone._session_factory = self._session_factory
        clone.__repository = self.__repository
         # enable=W0212
        return clone

    def get_by_id(self, id_key):
        ent = self._session.get_by_id(self.entity_class, id_key)
        if ent is None:
            try:
                ent = self.query().filter_by(id=id_key).one()
            except NoResultsException:
                pass
        if not ent is None \
           and not self._filter_spec is None \
           and not self._filter_spec.is_satisfied_by(ent):
            ent = None
        return ent

    def get_by_slug(self, slug):
        ent = self._session.get_by_slug(self.entity_class, slug)
        if ent is None:
            try:
                ent = self.query().filter_by(slug=slug).one()
            except NoResultsException:
                pass
        if not ent is None \
           and not self._filter_spec is None \
           and not self._filter_spec.is_satisfied_by(ent):
            ent = None
        return ent

    def add(self, data):
        self._session.add(self.entity_class, data)

    def remove(self, data):
        self._session.remove(self.entity_class, data)

    def update(self, data, target=None):
        return self._session.update(self.entity_class, data, target=target)

    def query(self, **options):
        return self._session.query(self.entity_class, **options)

    def sync_with_repository(self):
        self._session.flush()

    @property
    def expression_kind(self):
        return self._expression_kind

    def get_root_aggregate(self, rc):
        return self.__repository.get_aggregate(rc)

    def make_relationship_aggregate(self, relationship):
        """
        Returns a new relationship aggregate for the given relationship.

        :param relationship: Instance of
          :class:`everest.entities.relationship.DomainRelationship`.
        """
        if not self._session.IS_MANAGING_BACKREFERENCES:
            relationship.direction &= ~RELATIONSHIP_DIRECTIONS.REVERSE
        return RelationshipAggregate(self, relationship)

    @property
    def _session(self):
        return self._session_factory()


class RelationshipAggregate(Aggregate):
    """
    An aggregate that references a subset of a root aggregate defined through
    a relationship.
    """
    def __init__(self, root_aggregate, relationship):
        Aggregate.__init__(self)
        self._root_aggregate = root_aggregate
        self._relationship = relationship

    def get_by_id(self, id_key):
        ent = self._root_aggregate.get_by_id(id_key)
        if not ent is None and not self.filter.is_satisfied_by(ent):
            ent = None
        return ent

    def get_by_slug(self, slug):
        ent = self._root_aggregate.get_by_slug(slug)
        if not ent is None and not self.filter.is_satisfied_by(ent):
            ent = None
        return ent

    def query(self):
        return self._root_aggregate.query()

    def sync_with_repository(self):
        self._root_aggregate.sync_with_repository()

    @property
    def entity_class(self):
        return self._root_aggregate.entity_class

    @property
    def expression_kind(self):
        return self._root_aggregate.expression_kind

    def get_root_aggregate(self, rc):
        return self._root_aggregate.get_root_aggregate(rc)

    def add(self, entity):
        csc = self._relationship.descriptor.cascade
        add_to_root = csc & RELATION_OPERATIONS.ADD
        if add_to_root:
            self._root_aggregate.add(entity)
        self._relationship.add(entity, safe=True)

    def remove(self, entity):
        csc = self._relationship.descriptor.cascade
        remove_from_root = \
            csc & RELATION_OPERATIONS.REMOVE and not entity.id is None
        if remove_from_root:
            self._root_aggregate.remove(entity)
        self._relationship.remove(entity, safe=True)

    def update(self, entity, target=None):
        csc = self._relationship.descriptor.cascade
        if csc & RELATION_OPERATIONS.UPDATE:
            upd_entity = self._root_aggregate.update(entity, target=target)
        else:
            upd_entity = entity
        return upd_entity

    def _get_filter(self):
        # Overwrite to prepend relationship specification to filter spec.
        rel_spec = self._relationship.specification
        if not self._filter_spec is None:
            spec_fac = get_filter_specification_factory()
            filter_spec = spec_fac.create_conjunction(rel_spec,
                                                      self._filter_spec)
        else:
            filter_spec = rel_spec
        return filter_spec

    filter = property(_get_filter, Aggregate._set_filter)



"""
Static view.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jun 22, 2010.
"""
from pyramid.static import static_view
from pyramid.threadlocal import get_current_registry

__docformat__ = 'reStructuredText en'
__all__ = ['public_view',
           'PublicViewFactory',
           ]


class PublicViewFactory(object):

    PUBLIC_DIR = 'public_dir'
    PUBLIC_CACHE_MAX_AGE = 'public_cache_max_age'
    DEFAULT_CACHE_MAX_AGE = 3600

    def __init__(self):
        self.__static_view = None

    def __call__(self, context, request):
        if self.__static_view is None:
            self.__static_view = self._create_static_view()
        return self.__static_view(context, request)

    def _create_static_view(self):
        registry = get_current_registry()
        cache_max_age = int(registry.settings.get(self.PUBLIC_CACHE_MAX_AGE,
                                                  self.DEFAULT_CACHE_MAX_AGE))
        return static_view(registry.settings.get(self.PUBLIC_DIR),
                           cache_max_age=cache_max_age,
                           use_subpath=True)

public_view = PublicViewFactory()



"""
This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on May 31, 2012.
"""
import pytest

from everest.constants import DEFAULT_CASCADE
from everest.constants import RELATION_OPERATIONS
from everest.entities.attributes import get_domain_class_attribute
from everest.querying.specifications import AscendingOrderSpecification
from everest.querying.specifications import asc
from everest.querying.specifications import eq
from everest.querying.specifications import gt
from everest.repositories.memory.aggregate import MemoryAggregate
from everest.repositories.rdb.aggregate import RdbAggregate
from everest.tests.complete_app.entities import MyEntity
from everest.tests.complete_app.entities import MyEntityChild
from everest.tests.complete_app.entities import MyEntityParent
from everest.tests.complete_app.interfaces import IMyEntity
from everest.tests.complete_app.interfaces import IMyEntityChild
from everest.tests.complete_app.testing import create_entity


__docformat__ = 'reStructuredText en'
__all__ = []


class _TestRootAggregate(object):
    package_name = 'everest.tests.complete_app'
    agg_class = None

    def test_clone(self, entity_repo):
        agg = entity_repo.get_aggregate(IMyEntity)
        assert isinstance(agg, self.agg_class)
        agg_clone = agg.clone()
        for attr in ('entity_class', '_session_factory',
                     '_filter_spec', '_order_spec', '_slice_key',
                     '_RootAggregate__repository'):
            assert getattr(agg, attr) == getattr(agg_clone, attr)

    @pytest.mark.parametrize('ent_id0,ent_id1', [(0, 1)])
    def test_iterator_count(self, entity_repo, ent_id0, ent_id1):
        agg = entity_repo.get_aggregate(IMyEntity)
        with pytest.raises(StopIteration):
            next(agg.iterator())
        ent0 = create_entity(entity_id=ent_id0)
        agg.add(ent0)
        assert next(agg.iterator()) is ent0
        # Iterator heeds filtering.
        agg.filter = eq(id=ent_id1)
        with pytest.raises(StopIteration):
            next(agg.iterator())
        ent1 = create_entity(entity_id=ent_id1)
        agg.add(ent1)
        assert next(agg.iterator()) is ent1
        # Iterator heeds ordering.
        agg.filter = None
        agg.order = asc('id')
        assert isinstance(agg.order, AscendingOrderSpecification)
        assert next(agg.iterator()) is ent0
        # Iterator heeds slice.
        assert len(list(agg.iterator())) == 2
        agg.slice = slice(0, 1)
        assert len(list(agg.iterator())) == 1
        # Count ignores slice.
        assert agg.count() == 2
        # Count heeds filtering.
        agg.filter = eq(id=1)
        assert agg.count() == 1

    @pytest.mark.parametrize('ent_id', [0])
    def test_get_by_id_and_slug(self, entity_repo, ent_id):
        agg = entity_repo.get_aggregate(IMyEntity)
        ent = create_entity(entity_id=ent_id)
        agg.add(ent)
        assert agg.get_by_id(0)  is ent
        assert agg.get_by_slug('0') is ent
        assert agg.get_by_id(-1) is None
        assert agg.get_by_slug('-1') is None
        agg.filter = eq(id=1)
        assert agg.get_by_id(0) is None
        assert agg.get_by_slug('0') is None
        with pytest.raises(ValueError) as cm:
            agg.add(object())
        exp_msg = 'Invalid data type for traversal'
        assert str(cm.value).startswith(exp_msg)

    @pytest.mark.parametrize('data_ent0,data_ent1,data_ent2',
                             [((0, '222'), (1, '111'), (2, '000'))])
    def test_nested_attribute(self, entity_repo, data_ent0, data_ent1,
                              data_ent2):
        agg = entity_repo.get_aggregate(IMyEntity)
        ent0 = create_entity(entity_id=data_ent0[0])
        ent0.parent.text_ent = data_ent0[1]
        ent1 = create_entity(entity_id=data_ent1[0])
        ent1.parent.text_ent = data_ent1[1]
        ent2 = create_entity(entity_id=data_ent2[0])
        ent2.parent.text_ent = data_ent2[1]
        agg.add(ent0)
        agg.add(ent1)
        agg.add(ent2)
        assert len(list(agg.iterator())) == 3
        agg.filter = eq(**{'parent.text_ent':'222'})
        assert len(list(agg.iterator())) == 1
        agg.filter = None
        assert len(list(agg.iterator())) == 3
        agg.order = asc('parent.text_ent')
        assert next(agg.iterator()) is ent2
        # With nested filter and order.
        agg.filter = gt(**{'parent.text_ent':'000'})
        assert next(agg.iterator()) is ent1
        # With nested filter, order, and slice.
        agg.slice = slice(1, 2)
        assert next(agg.iterator()) is ent0

    @pytest.mark.parametrize('ent_id', [0])
    def test_add_remove(self, entity_repo, ent_id):
        agg = entity_repo.get_aggregate(IMyEntity)
        ent = create_entity(entity_id=ent_id)
        agg.add(ent)
        assert len(list(agg.iterator())) == 1
        agg.remove(ent)
        assert len(list(agg.iterator())) == 0


class TestMemoryRootAggregate(_TestRootAggregate):
    config_file_name = 'configure_no_rdb.zcml'
    agg_class = MemoryAggregate


@pytest.mark.usefixtures('rdb')
class TestRdbRootAggregate(_TestRootAggregate):
    agg_class = RdbAggregate


class _TestRelationshipAggregate(object):
    package_name = 'everest.tests.complete_app'

    def _make_rel_agg(self, entity_repo, entity, attr_name=None):
        if attr_name is None:
            attr_name = 'children'
        attr = get_domain_class_attribute(entity, attr_name)
        rel = attr.make_relationship(entity)
        child_agg = entity_repo.get_aggregate(attr.attr_type)
        return child_agg.make_relationship_aggregate(rel)

    def _make_child(self, child_agg, child_id=0):
        new_parent = MyEntityParent()
        new_ent = MyEntity()
        new_ent.parent = new_parent
        new_child = MyEntityChild()
        new_ent.children.append(new_child)
        if new_child.parent is None:
            new_child.parent = new_ent
        child_agg.add(new_child)
        new_parent.id = child_id
        new_ent.id = child_id
        new_child.id = child_id
        return new_child

    def test_basics(self, entity_repo):
        agg = entity_repo.get_aggregate(IMyEntity)
        child_agg = entity_repo.get_aggregate(IMyEntityChild)
        new_child0 = self._make_child(child_agg)
        new_parent1 = MyEntityParent()
        new_ent1 = MyEntity()
        new_ent1.parent = new_parent1
        new_child1 = MyEntityChild()
        new_child1.parent = new_ent1
        child_rel_agg = self._make_rel_agg(entity_repo, new_ent1)
        assert len(list(child_agg.iterator())) == 1
        assert len(list(agg.iterator())) == 1
        assert len(list(child_rel_agg.iterator())) == 0
        # Adding to a relationship aggregate .....
        child_rel_agg.add(new_child1)
        # ....... adds to root aggregates:
        assert len(list(child_agg.iterator())) == 2
        # ....... adds (along the cascade) to parent root aggregate:
        assert len(list(agg.iterator())) == 2
        # ....... appends to children:
        assert new_ent1.children == [new_child1]
        # get by ID and slug, filtering.
        assert child_rel_agg.get_by_id(new_child1.id).id == new_child1.id
        assert \
            child_rel_agg.get_by_slug(new_child1.slug).slug == new_child1.slug
        child_rel_agg.filter = eq(id=2)
        assert child_rel_agg.get_by_id(new_child1.id) is None
        assert child_rel_agg.get_by_slug(new_child1.slug) is None
        # update.
        upd_child0 = MyEntityChild(id=0)
        txt = 'FROBNIC'
        upd_child0.text = txt
        child_rel_agg.update(upd_child0)
        assert new_child0.text == txt
        # FIXME: The RDB backend behaves different from the memory backend
        #        here (parent is not None unless the child is also removed
        #        from the children container).
#        assert new_child0.parent is None

    def test_update_cascade(self, entity_repo, monkeypatch):
        csc = DEFAULT_CASCADE & ~RELATION_OPERATIONS.UPDATE
        child_agg = entity_repo.get_aggregate(IMyEntityChild)
        new_child = self._make_child(child_agg)
        child_rel_agg = self._make_rel_agg(entity_repo, new_child.parent)
        children_attr = get_domain_class_attribute(MyEntity, 'children')
        monkeypatch.setattr(children_attr, 'cascade', csc)
        upd_child = MyEntityChild(id=0)
        txt = 'FROBNIC'
        upd_child.text = txt
        child_rel_agg.update(upd_child)
        assert new_child.text != txt
        assert not new_child.parent is None

    def test_add_one_to_one(self, entity_repo):
        new_parent1 = MyEntityParent(id=1)
        new_ent1 = MyEntity(id=1)
        parent_rel_agg = self._make_rel_agg(entity_repo, new_ent1, 'parent')
        assert new_ent1.parent is None
        parent_rel_agg.add(new_parent1)
        assert new_ent1.parent == new_parent1

    def test_delete_cascade(self, entity_repo, monkeypatch):
        new_parent1 = MyEntityParent()
        new_ent1 = MyEntity()
        new_ent1.parent = new_parent1
        new_child1 = MyEntityChild()
        new_child1.parent = new_ent1
        child_rel_agg = self._make_rel_agg(entity_repo, new_ent1)
        child_rel_agg.add(new_child1)
        new_parent1.id = 1
        new_ent1.id = 1
        new_child1.id = 1
        agg = entity_repo.get_aggregate(IMyEntity)
        child_agg = entity_repo.get_aggregate(IMyEntityChild)
        assert len(list(child_agg.iterator())) == 1
        assert len(list(agg.iterator())) == 1
        assert new_ent1.children == [new_child1]
        assert new_child1.parent == new_ent1
        csc = DEFAULT_CASCADE | RELATION_OPERATIONS.REMOVE
        children_attr = get_domain_class_attribute(MyEntity, 'children')
        parent_attr = get_domain_class_attribute(MyEntityChild, 'parent')
        monkeypatch.setattr(children_attr, 'cascade', csc)
        monkeypatch.setattr(parent_attr, 'cascade', csc)
        child_rel_agg.remove(new_child1)
        assert new_ent1.children == []
        assert new_child1.parent is None
        assert len(list(child_agg.iterator())) == 0
        if self.__class__.__name__.startswith('TestMemory'):
            # FIXME: Transparent modification of RDB mapper cascades
            #        does not work yet.
            assert len(list(agg.iterator())) == 0
        assert len(list(child_rel_agg.iterator())) == 0


class TestMemoryRelationshipAggregate(_TestRelationshipAggregate):
    config_file_name = 'configure_no_rdb.zcml'


@pytest.mark.usefixtures('rdb')
class TestRdbRelationshipAggregate(_TestRelationshipAggregate):
    pass



"""
Root factory.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 21, 2011.
"""
from everest.resources.utils import get_service

__docformat__ = 'reStructuredText en'
__all__ = ['RootFactory',
           ]


class RootFactory(object):
    """
    Factory for the object tree root (i.e., the service object).
    """
    def __init__(self):
        self.__root = None

    def __call__(self, request):
        if self.__root is None:
            # Start the service.
            self.__root = get_service()
            self.__root.start()
        return self.__root




"""
Order CQL expression parser.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Jul 5, 2011.
"""
from everest.entities.utils import identifier_from_slug
from everest.querying.operators import ASCENDING
from everest.querying.operators import DESCENDING
from everest.querying.specifications import asc
from everest.querying.specifications import desc
from pyparsing import Combine
from pyparsing import Group
from pyparsing import Literal
from pyparsing import Word
from pyparsing import ZeroOrMore
from pyparsing import alphas
from pyparsing import oneOf
from pyparsing import opAssoc
from pyparsing import operatorPrecedence

__docformat__ = 'reStructuredText en'
__all__ = ['parse_order',
           ]


colon = Literal(':')
tilde = Literal('~')
attribute = Word(alphas, alphas + '-')
identifier = Combine(attribute + ZeroOrMore('.' + attribute)
                     ).setName('identifier')
direction = oneOf("asc desc").setName('direction')

BINARY = 2

class CriterionConverter(object):
    spec_map = {ASCENDING.name:asc,
                DESCENDING.name:desc}

    @classmethod
    def convert(cls, seq):
        crit = seq[0]
        name = cls.__prepare_identifier(crit.name)
        op_name = cls.__prepare_identifier(crit.operator)
        return cls.__make_spec(name, op_name)

    @classmethod
    def __prepare_identifier(cls, name):
        return identifier_from_slug(name)

    @classmethod
    def __make_spec(cls, name, op_name):
        spec_gen = cls.spec_map[op_name]
        return spec_gen(name)


def convert_junction(seq):
    left_spec = seq[0][0]
    right_spec = seq[0][-1]
    return left_spec & right_spec


criterion = Group(identifier('name') + colon.suppress() + \
                  direction('operator')).setResultsName('order')
criterion.setParseAction(CriterionConverter.convert)
criteria = \
    operatorPrecedence(criterion,
                       [
                        (tilde, BINARY, opAssoc.LEFT, convert_junction)
                        ])

order = criteria | criterion


def parse_order(criteria_string):
    """
    Parses the given order criteria string.
    """
    return order.parseString(criteria_string)[0]



"""
Resource base classes.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Nov 3, 2011.
"""
from everest.constants import RESOURCE_ATTRIBUTE_KINDS
from everest.entities.utils import get_entity_class
from everest.entities.utils import identifier_from_slug
from everest.entities.utils import slug_from_identifier
from everest.querying.base import SpecificationExpressionHolder
from everest.querying.interfaces import ISpecificationVisitor
from everest.querying.specifications import AscendingOrderSpecification
from everest.querying.utils import get_filter_specification_factory
from everest.resources.attributes import ResourceAttributeControllerMixin
from everest.resources.attributes import get_resource_class_attribute
from everest.resources.descriptors import terminal_attribute
from everest.resources.interfaces import ICollectionResource
from everest.resources.interfaces import IMemberResource
from everest.resources.interfaces import IResource
from everest.resources.link import Link
from everest.resources.utils import as_member
from everest.resources.utils import get_collection_class
from everest.resources.utils import get_member_class
from pyramid.security import Allow
from pyramid.security import Authenticated
from pyramid.traversal import model_path
from zope.interface import implementer # pylint: disable=E0611,F0401
import uuid

__docformat__ = "reStructuredText en"
__all__ = ['Collection',
           'Member',
           'Resource',
           'ResourceToEntityFilterSpecificationVisitor',
           'ResourceToEntityOrderSpecificationVisitor',
           'ResourceToEntitySpecificationVisitor',
           ]


@implementer(IResource)
class Resource(object):
    """
    Abstract base class for all resources.
    """

    #: Authentication specifier. Override as needed.
    __acl__ = [
        (Allow, Authenticated, 'view'),
        (Allow, Authenticated, 'create'),
        (Allow, Authenticated, 'update'),
        (Allow, Authenticated, 'delete'),
        ]
    #: The parent of this resource. This is `None` for the service resource.
    __parent__ = None
    #: The name of the resource. This has to be unique within the parent.
    __name__ = None
    #: The relation identifier to show in links to this resource. Needs to
    #: be specified in derived classes.
    relation = None
    #: Descriptive title for this resource.
    title = ''
    #: Detailed description of this resource.
    description = ''

    def __init__(self, relationship=None):
        """
        Constructor:
        """
        if self.__class__ is Resource:
            raise NotImplementedError('Abstract class')
        if self.__class__.relation is None:
            raise ValueError('Resource classes must have a relation '
                             'attribute.')
        #: A relationship to some other resource. Makes this resource a
        #: "nested" resource.
        self._relationship = relationship
        self.is_nested = not relationship is None
        #: A set of links to other resources.
        self.links = set()

    def add_link(self, link):
        """
        Adds a link to another resource.

        :param link: Resource link.
        :type link: :class:`everest.resources.base.Link`
        """
        self.links.add(link)

    @property
    def path(self):
        """
        Returns the path to this resource in the tree of resources.
        """
        return model_path(self)

    @property
    def urn(self):
        """
        Returns the URN for this resource (globally unique identifier).
        """
        return uuid.uuid5(uuid.NAMESPACE_URL, self.path).urn

    @property
    def has_parent(self):
        """
        Checks if this resource has a parent.
        """
        return not self.__parent__ is None


@implementer(IMemberResource)
class Member(ResourceAttributeControllerMixin, Resource):
    """
    Base class for all member resources.
    """

    id = terminal_attribute(int, 'id')

    def __init__(self, entity, name=None, relationship=None):
        """
        Constructor:

        :param str name: Unique name of the member within the collection
        :param entity: Associated entity (domain object).
        :type entity: Object implementing an interface derived from
                :class:`everest.entities.interfaces.IEntity`.
        """
        if self.__class__ is Member:
            raise NotImplementedError('Abstract class')
        if not isinstance(entity, get_entity_class(self)):
            raise ValueError(
                    'Invalid entity class "%s" for %s resource class.'
                    % (entity.__class__.__name__, self.__class__.__name__))
        super(Member, self).__init__(relationship=relationship)
        self.__entity = entity
        self.__name = name
        # Add the rel="self" link.
        self.add_link(Link(self, "self"))

    def _get__name__(self):
        # The name of a member resource defaults to the slug of the underlying
        # entity.
        return self.__name or self.__entity.slug

    def _set__name__(self, name):
        self.__name = name

    __name__ = property(_get__name__, _set__name__)

    @classmethod
    def create_from_entity(cls, entity):
        """
        Class factory method creating a new resource from the given entity.
        """
        return cls(entity)

    def get_entity(self):
        """
        Returns the entity this resource manages.

        :return: Object implementing
            :class:`everest.entities.interfaces.IEntity`.
        """
        return self.__entity

    def update(self, data):
        """
        Updates this member from the given data.

        See :method:`Collection.update`.
        """
        self.__parent__.update(data, target=self)

    def __getitem__(self, item):
        ident = identifier_from_slug(item)
        attr = get_resource_class_attribute(self.__class__, ident)
        if attr is None:
            raise KeyError('%s' % ident)
        return getattr(self, ident)

    def __hash__(self):
        """
        Hash value, based on the resource class and the name.
        """
        return hash((self.__class__, self.__name__))

    def __eq__(self, other):
        """
        Equality operator. Two members compare equal if they belong to the
        same class and have the same name.
        """
        return (isinstance(other, self.__class__) and
                self.__name__ == other.__name__)

    def __ne__(self, other):
        """
        Inequality operator.
        """
        return not (self == other)

    def __str__(self):
        return "%s(id: %s, name: %s)" \
               % (self.__class__.__name__, self.id, self.__name__)

    @property
    def is_root_member(self):
        # The parent might be a member, so we use getattr with a default.
        return not self.__parent__ is None \
               and getattr(self.__parent__, 'is_root_collection', False)

    @classmethod
    def as_related_member(cls, entity, relationship):
        """
        Creates a new relationship member with the relationship's relator
        as a parent.
        """
        rel_mb = cls.create_from_entity(entity)
        rel_mb._relationship = relationship # pylint: disable=W0212
        rel_mb.__parent__ = relationship.relator
        return rel_mb


@implementer(ICollectionResource)
class Collection(Resource):
    """
    This is an abstract base class for all resource collections.
    A collection is a set of member resources which can be filtered, sorted,
    and sliced.
    """
    #: The title of the collection.
    title = None
    #: The name for the root collection (used as URL path to the root
    #: collection inside the service).
    root_name = None
    #: A description of the collection.
    description = ''
    #: The default order of the collection's members.
    default_order = AscendingOrderSpecification('id')
    #: The page size default for this collection.
    default_limit = 100
    #: The default maximum page size limit for this collection. Unless
    #: this is set in derived classes, no limit is enforced (i.e., the
    #: default maximum limit is None).
    max_limit = None

    def __init__(self, aggregate, name=None, relationship=None):
        """
        Constructor.

        :param str name: Name of the collection.
        :param aggregate: Associated aggregate.
        :type aggregate: :class:`everest.entities.aggregates.Aggregate` -
                an object implementing an interface derived from
                :class:`everest.entities.interfaces.IAggregate`.
        """
        if self.__class__ is Collection:
            raise NotImplementedError('Abstract class')
        Resource.__init__(self, relationship=relationship)
        if self.title is None:
            raise ValueError('Collection must have a title.')
        if name is None:
            name = self.root_name
        self.__name__ = name
        #: The filter specification for this resource. Attribute names in
        #: this specification are relative to the resource.
        self._filter_spec = None
        #: The order specification for this resource. Attribute names in
        #: this specification are relative to the resource.
        self._order_spec = None
        # The underlying aggregate.
        self.__aggregate = aggregate

    @classmethod
    def create_from_aggregate(cls, aggregate, relationship=None):
        """
        Creates a new collection from the given aggregate.

        :param aggregate: Aggregate containing the entities exposed by this
            collection resource.
        :param relationship: Resource relationship. If given, the root
            aggregate is converted to a relationship aggregate and the
            relationship is passed on to the collection class constructor.
        :type aggregate: :class:`everest.entities.aggregates.RootAggregate`
        """
        if not relationship is None:
            aggregate = aggregate.make_relationship_aggregate(
                                            relationship.domain_relationship)
        return cls(aggregate, relationship=relationship)

    def get_aggregate(self):
        """
        Returns the aggregate underlying this collection.

        :returns: Object implementing
            :class:`everest.entities.interfaces.IAggregate`.
        """
        return self.__aggregate

    def create_member(self, entity):
        """
        Creates a new member resource from the given entity and adds it to
        this collection.
        """
        member = as_member(entity)
        self.add(member)
        return member

    def __len__(self):
        """
        Returns the size (count) of the collection.
        """
        return self.__aggregate.count()

    def __getitem__(self, key):
        """
        Gets a member (by name).

        :param key: Name of the member
        :type key: :class:`string` or :class:`unicode`
        :raises: :class:`everest.exceptions.MultipleResultsException` if more
          than one member is found for the given key value.
        :raises: KeyError If no entity is found for the given key.
        :returns: Object implementing
          :class:`everest.resources.interfaces.IMemberResource`.
        """
        ent = self.__aggregate.get_by_slug(key)
        if ent is None:
            raise KeyError(key)
        rc = as_member(ent, parent=self)
        return rc

    def __iter__(self):
        """
        Returns an iterator over the (possibly filtered and ordered)
        collection.
        """
        for obj in self.__aggregate.iterator():
            rc = as_member(obj, parent=self)
            yield rc

    def __contains__(self, member):
        """
        Checks if this collection contains the given member.

        :returns: `False` if a lookup of the ID of the given member returns
            `None` or if the ID is `None`; else, `True`.
        """
        return not (member.id is None \
                    or self.__aggregate.get_by_id(member.id) is None)

    def __str__(self):
        return "<%s name:%s parent:%s>" % (self.__class__.__name__,
                                           self.__name__, self.__parent__)

    def __repr__(self):
        return self.__str__()

    def add(self, member):
        """
        Adds the given member to this collection.

        :param member: Member to add.
        :type member: Object implementing
                    :class:`everest.resources.interfaces.IMemberResource`
        :raise ValueError: if a member with the same name exists
        """
        if IMemberResource.providedBy(member): #pylint: disable=E1101
            member.__parent__ = self
            data = member.get_entity()
        else:
            data = member
        self.__aggregate.add(data)

    def remove(self, member):
        """
        Removes the given member from this collection.

        :param member: Member to remove.
        :type member: Object implementing
                    :class:`everest.resources.interfaces.IMemberResource`
        :raise ValueError: if the member can not be found in this collection
        """
        is_member = IMemberResource.providedBy(member) #pylint: disable=E1101
        if is_member:
            data = member.get_entity()
        else:
            data = member
        self.__aggregate.remove(data)
        if is_member:
            member.__parent__ = None

    def get(self, key, default=None):
        """
        Returns a member for the given key or the given default value if no
        match was found in the collection.
        """
        try:
            rc = self.__getitem__(key)
        except KeyError:
            rc = default
        return rc

    def update(self, data, target=None):
        """
        Updates this collection from the given data.

        :param data: Any object that can be adapted to
          :class:`everest.interfaces.IDataTraversalProxyAdapter`.
        :returns: New updated member.
        """
        if not target is None:
            target = target.get_entity()
        updated_entity = self.__aggregate.update(data, target=target)
        return as_member(updated_entity, parent=self)

    def _get_filter(self):
        if self._relationship is None:
            filter_spec = self._filter_spec
        else:
            rel_spec = self._relationship.specification
            if self._filter_spec is None:
                filter_spec = rel_spec
            else:
                spec_fac = get_filter_specification_factory()
                filter_spec = spec_fac.create_conjunction(rel_spec,
                                                          self._filter_spec)
        return filter_spec

    def _set_filter(self, filter_spec):
        if not filter_spec is None:
            # Translate to entity filter expression before passing on to the
            # aggregate.
            visitor = ResourceToEntityFilterSpecificationVisitor(
                                                    get_member_class(self))
            filter_spec.accept(visitor)
            self.__aggregate.filter = visitor.expression
        else:
            self.__aggregate.filter = None
        self._filter_spec = filter_spec

    filter = property(_get_filter, _set_filter)

    def _get_order(self):
        return self._order_spec

    def _set_order(self, order_spec):
        if not order_spec is None:
            # Translate to entity order expression before passing on to the
            # aggregate.
            visitor = ResourceToEntityOrderSpecificationVisitor(
                                                    get_member_class(self))
            order_spec.accept(visitor)
            self.__aggregate.order = visitor.expression
        else:
            self.__aggregate.order = None
        self._order_spec = order_spec

    order = property(_get_order, _set_order)

    def _get_slice(self):
        return self.__aggregate.slice

    def _set_slice(self, slice_key):
        self.__aggregate.slice = slice_key

    slice = property(_get_slice, _set_slice)

    def clone(self):
        """
        Returns a clone of this collection.
        """
        agg = self.__aggregate.clone()
        clone = self.create_from_aggregate(agg)
        # Pass filter and order specs explicitly (may differ from the ones
        # at the aggregate level).
        # pylint: disable=W0212
        clone._filter_spec = self._filter_spec
        clone._order_spec = self._order_spec
        # pylint: enable=W0212
        clone.__parent__ = self.__parent__
        return clone

    @property
    def is_root_collection(self):
        return self._relationship is None and not self.__parent__ is None

    def get_root_collection(self, rc):
        """
        Returns a root collection for the given resource.

        The root collection is created using a root aggregate fetched from
        the same repository that was used to create the root aggregate for
        this collection.
        """
        root_agg = self.__aggregate.get_root_aggregate(rc)
        coll_cls = get_collection_class(rc)
        return coll_cls.create_from_aggregate(root_agg)

    @classmethod
    def as_related_collection(cls, aggregate, relationship):
        """
        Creates a new relationship collection with a relationship aggregate
        and the relationship's relator as a parent.
        """
        rel_coll = cls.create_from_aggregate(aggregate,
                                             relationship=relationship)
        # The member at the origin of the relationship is the parent.
        rel_coll.__parent__ = relationship.relator
        # Set the collection's name to the descriptor's resource
        # attribute name.
        rel_coll.__name__ = \
            slug_from_identifier(relationship.descriptor.resource_attr)
        return rel_coll


@implementer(ISpecificationVisitor)
class ResourceToEntitySpecificationVisitor(SpecificationExpressionHolder):
    """
    Base class for specification visitors that convert resource to entity
    attribute names.
    """

    def __init__(self, rc_class):
        SpecificationExpressionHolder.__init__(self)
        self.__rc_class = rc_class

    def visit_nullary(self, spec):
        entity_attr_name = self.__convert_to_entity_attr(spec.attr_name)
        new_spec = self._make_new_spec(entity_attr_name, spec)
        self._push(new_spec)

    def visit_unary(self, spec):
        last = self._pop()
        new_spec = spec.__class__(last)
        self._push(new_spec)

    def visit_binary(self, spec):
        right = self._pop()
        left = self._pop()
        new_spec = spec.__class__(left, right)
        self._push(new_spec)

    def __convert_to_entity_attr(self, rc_attr_name):
        entity_attr_tokens = []
        rc_class = self.__rc_class
        for rc_attr_token in rc_attr_name.split('.'):
            rc_attr = get_resource_class_attribute(rc_class, rc_attr_token)
            if rc_attr is None:
                raise AttributeError('%s resource does not have an attribute '
                                     '"%s".'
                                     % (rc_class.__name__, rc_attr_name))
            ent_attr_name = rc_attr.entity_attr
            if ent_attr_name is None:
                raise ValueError('Resource attribute "%s" does not have a '
                                 'corresponding entity attribute.'
                                 % rc_attr.resource_attr)
            if rc_attr.kind != RESOURCE_ATTRIBUTE_KINDS.TERMINAL:
                # Look up the member class for the specified member or
                # collection resource interface.
                rc_class = get_member_class(rc_attr.attr_type)
            entity_attr_tokens.append(ent_attr_name)
        return '.'.join(entity_attr_tokens)

    def _make_new_spec(self, new_attr_name, old_spec):
        raise NotImplementedError('Abstract method.')


class ResourceToEntityFilterSpecificationVisitor(
                                        ResourceToEntitySpecificationVisitor):
    """
    Filter specification visitor that converts resource attribute names to
    entity attribute names.
    """
    def _make_new_spec(self, new_attr_name, old_spec):
        return old_spec.__class__(new_attr_name, old_spec.attr_value)


class ResourceToEntityOrderSpecificationVisitor(
                                        ResourceToEntitySpecificationVisitor):
    """
    Order specification visitor that converts resource attribute names to
    entity attribute names.
    """
    def _make_new_spec(self, new_attr_name, old_spec):
        return old_spec.__class__(new_attr_name)



"""
View base classes.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Oct 7, 2011.j
"""
import logging
import re

from pyramid.httpexceptions import HTTPBadRequest
from pyramid.httpexceptions import HTTPConflict
from pyramid.httpexceptions import HTTPError
from pyramid.httpexceptions import HTTPInternalServerError # pylint: disable=F0401
from pyramid.httpexceptions import HTTPNotAcceptable
from pyramid.httpexceptions import HTTPOk
from pyramid.httpexceptions import HTTPTemporaryRedirect # pylint: disable=F0401
from pyramid.httpexceptions import HTTPUnsupportedMediaType
from pyramid.interfaces import IResponse
from pyramid.threadlocal import get_current_request

from everest.messaging import UserMessageChecker
from everest.messaging import UserMessageHandlingContextManager
from everest.mime import CsvMime
from everest.mime import TextPlainMime
from everest.mime import get_registered_mime_strings
from everest.mime import get_registered_mime_type_for_name
from everest.mime import get_registered_mime_type_for_string
from everest.representers.utils import UpdatedRepresenterConfigurationContext
from everest.representers.utils import as_representer
from everest.resources.system import UserMessageMember
from everest.resources.utils import resource_to_url
from everest.url import UrlPartsConverter
from everest.utils import get_traceback
from everest.views.interfaces import IResourceView
from zope.interface import implementer # pylint: disable=E0611,F0401


__docformat__ = "reStructuredText en"
__all__ = ['GetResourceView',
           'HttpWarningResubmit',
           'ModifyingResourceView',
           'RepresentingResourceView',
           'ResourceView',
           'WarnAndResubmitExecutor',
           'WarnAndResubmitUserMessageChecker',
           ]


class WarnAndResubmitExecutor(object):
    """
    Executes a callable within a user message handling context that uses a
    :class:`WarnAndResutbmitUserMessageChecker` as checker.
    """
    def __init__(self, func):
        """
        Constructor.

        :param func: Callable to execute.
        """
        self.__func = func
        self.__checker = None

    def __call__(self, *args, **kw):
        """
        Runs the function passed to this executor with the given positional
        and keyword arguments within a user message handling context.

        Returns the result of the execution or a 307 Temporary Redirect
        response in case the user message checker voted to stop further
        processing.
        """
        self.__checker = WarnAndResubmitUserMessageChecker()
        with UserMessageHandlingContextManager(self.__checker):
            result = self.__func(*args, **kw)
        if not self.__checker.vote is True:
            result = self.__checker.create_307_response()
        return result

    @property
    def do_continue(self):
        """
        Returns `True` if the checker the executor has run and the checker
        did not vote to stop further processing.
        """
        return not self.__checker is None and self.__checker.vote is True


class HttpWarningResubmit(HTTPTemporaryRedirect): # no __init__ pylint: disable=W0232
    """
    Special 307 HTTP Temporary Redirect exception which transports a URL
    at which the user may resubmit the request with the warning suppressed.
    """
    explanation = 'Your request triggered warnings. You may resubmit ' \
                  'the request under the given location to ignore ' \
                  'the warnings.'
    template = '%(explanation)s\r\n' \
               'Location: <a href="%(location)s">%(location)s</a>\r\n' \
               'Warning UserMessage: %(detail)s\r\n' \
               '<!-- %(comment)s -->'


@implementer(IResourceView)
class ResourceView(object):
    """
    Abstract base class for all resource views.

    Resource views know how to handle a number of things that can go wrong
    in a REST request.
    """
    def __init__(self, context, request):
        if self.__class__ is ResourceView:
            raise NotImplementedError('Abstract class')
        #: View logger (qualname "everest.views").
        self._logger = logging.getLogger('everest.views')
        #: The contetxt of the view.
        self.__context = context
        #: The request for the view.
        self.__request = request

    @property
    def context(self):
        return self.__context

    @property
    def request(self):
        return self.__request

    def __call__(self):
        raise NotImplementedError('Abstract method.')

    def _handle_unknown_exception(self, message, traceback):
        """
        Handles requests that triggered an unknown exception.

        Respond with a 500 "Internal Server Error".
        """
        self._logger.error('Request errors\n'
                           'Error message: %s\nTraceback:%s' %
                           (message, traceback))
        http_exc = HTTPInternalServerError(message)
        return self.request.get_response(http_exc)


class RepresentingResourceView(ResourceView): # still abstract pylint: disable=W0223
    """
    A resource view with an associated representer.
    """
    def __init__(self, context, request,
                 default_content_type=None,
                 default_response_content_type=None, convert_response=True,
                 enable_messaging=False):
        if self.__class__ is RepresentingResourceView:
            raise NotImplementedError('Abstract class')
        ResourceView.__init__(self, context, request)
        #: Flag indicating if the response body should be converted using
        #: the representer associated with this view.
        self._convert_response = convert_response
        if default_content_type is None:
            # FIXME: make this configurable.
            default_content_type = CsvMime
        #: The default content type for this view.
        self._default_content_type = default_content_type
        #: The default content type for the response.
        self._default_response_content_type = default_response_content_type
        #: Flag indicating if a messaging context should be used when
        #: processing calls into this view.
        self._enable_messaging = enable_messaging

    def _get_response_mime_type(self):
        """
        Returns the reponse MIME type for this view.

        :raises: :class:`pyramid.httpexceptions.HTTPNotAcceptable` if the
          MIME content type(s) the client specified can not be handled by
          the view.
        """
        view_name = self.request.view_name
        if view_name != '':
            mime_type = get_registered_mime_type_for_name(view_name)
        else:
            mime_type = None
            acc = None
            for acc in self.request.accept:
                if acc == '*/*':
                    # The client does not care; use the default.
                    mime_type = self.__get_default_response_mime_type()
                    break
                try:
                    mime_type = \
                            get_registered_mime_type_for_string(acc.lower())
                except KeyError:
                    pass
                else:
                    break
            if mime_type is None:
                if not acc is None:
                    # The client specified a MIME type we can not handle; this
                    # is a 406 exception. We supply allowed MIME content
                    # types in the body of the response.
                    headers = \
                        [('Location', self.request.path_url),
                         ('Content-Type', TextPlainMime.mime_type_string),
                         ]
                    mime_strings = get_registered_mime_strings()
                    exc = HTTPNotAcceptable('Requested MIME content type(s) '
                                            'not acceptable.',
                                            body=','.join(mime_strings),
                                            headers=headers)
                    raise exc
                mime_type = self.__get_default_response_mime_type()
        return mime_type

    def _get_response_representer(self):
        """
        Creates a representer for this view.

        :returns: :class:`everest.representers.base.ResourceRepresenter`
        """
        mime_type = self._get_response_mime_type()
        return as_representer(self.context, mime_type)

    def _get_result(self, resource):
        """
        Converts the given resource to a result to be returned from the view.
        Unless a custom renderer is employed, this will involve creating
        a representer and using it to convert the resource to a string.

        :param resource: Resource to convert.
        :type resource: Object implementing
          :class:`evererst.interfaces.IResource`.
        :returns: :class:`pyramid.reposnse.Response` object or a dictionary
          with a single key "context" mapped to the given resource (to be
          passed on to a custom renderer).
        """
        if self._convert_response:
            self._update_response_body(resource)
            result = self.request.response
        else:
            result = dict(context=resource)
        return result

    def _update_response_body(self, resource):
        """
        Creates a representer and updates the response body with the byte
        representation created for the given resource.
        """
        rpr = self._get_response_representer()
        # Set content type and body of the response.
        self.request.response.content_type = \
                                rpr.content_type.mime_type_string
        rpr_body = rpr.to_bytes(resource)
        self.request.response.body = rpr_body

    def _update_response_location_header(self, resource):
        """
        Adds a new or replaces an existing Location header to the response
        headers pointing to the URL of the given resource.
        """
        location = resource_to_url(resource, request=self.request)
        loc_hdr = ('Location', location)
        hdr_names = [hdr[0].upper() for hdr in self.request.response.headerlist]
        try:
            idx = hdr_names.index('LOCATION')
        except ValueError:
            self.request.response.headerlist.append(loc_hdr)
        else:
            # Replace existing location header.
            # FIXME: It is not clear under which conditions this happens, so
            #        we do not have a test for it yet.
            self.request.response.headerlist[idx] = loc_hdr # pragma: no cover

    def __get_default_response_mime_type(self):
        if not self._default_response_content_type is None:
            mime_type = self._default_response_content_type
        else:
            mime_type = self._default_content_type
        return mime_type


class GetResourceView(RepresentingResourceView): # still abstract pylint: disable=W0223
    """
    Abstract base class for all resource views processing GET requests.
    """
    def __init__(self, resource, request, **kw):
        if self.__class__ is GetResourceView:
            raise NotImplementedError('Abstract class')
        # Messaging is disabled by default for GET views.
        if kw.get('enable_messaging') is None:
            kw['enable_messaging'] = False
        RepresentingResourceView.__init__(self, resource, request, **kw)

    def __call__(self):
        self._logger.debug('Request URL: %s.', self.request.url)
        try:
            if self._enable_messaging:
                prep_executor = \
                    WarnAndResubmitExecutor(self._prepare_resource)
                data = prep_executor()
                do_continue = prep_executor.do_continue
            else:
                data = self._prepare_resource()
                do_continue = not IResponse.providedBy(data) # pylint: disable=E1101
            if do_continue:
                # Return a response to bypass Pyramid rendering.
                if self._enable_messaging:
                    res_executor = WarnAndResubmitExecutor(self._get_result)
                    result = res_executor(data)
                else:
                    result = self._get_result(data)
            else:
                result = data
        except HTTPError as http_exc:
            result = self.request.get_response(http_exc)
        except Exception as err: # catch Exception pylint: disable=W0703
            result = self._handle_unknown_exception(str(err),
                                                    get_traceback())
        return result

    def _prepare_resource(self):
        raise NotImplementedError('Abstract method.')

    def _update_response_body(self, resource):
        """
        Extends the base class method with links options processing.
        """
        links_options = self.__configure_refs()
        if not links_options is None:
            with UpdatedRepresenterConfigurationContext(
                                        type(self.context),
                                        self._get_response_mime_type(),
                                        attribute_options=links_options):
                RepresentingResourceView._update_response_body(self, resource)
        else:
            RepresentingResourceView._update_response_body(self, resource)

    def __configure_refs(self):
        refs_options_string = self.request.params.get('refs')
        if not refs_options_string is None:
            links_options = \
                UrlPartsConverter.make_refs_options(refs_options_string)
        else:
            links_options = None
        return links_options


class ModifyingResourceView(RepresentingResourceView): # still abstract pylint: disable=W0223
    """
    Abstract base class for all modifying member views.
    """
    def __init__(self, resource, request, **kw):
        if self.__class__ is ModifyingResourceView:
            raise NotImplementedError('Abstract class')
        # Messaging is enabled by default for modifying views.
        if kw.get('enable_messaging') is None:
            kw['enable_messaging'] = True
        RepresentingResourceView.__init__(self, resource, request, **kw)

    def __call__(self):
        self._logger.debug('%s request received on %s.',
                           self.request.method, self.request.url)
        self._logger.debug('Request body: %s', self.request.body,
                           extra=dict(output_limit=500))
        if len(self.request.body) == 0:
            # Empty body - return 400 Bad Request.
            result = self._handle_empty_body()
        else:
            try:
                if self._enable_messaging:
                    extract_executor = \
                        WarnAndResubmitExecutor(self._extract_request_data)
                    data = extract_executor()
                    do_continue = extract_executor.do_continue
                else:
                    data = self._extract_request_data()
                    do_continue = not IResponse.providedBy(data) # pylint: disable=E1101
                if do_continue:
                    if self._enable_messaging:
                        process_executor = \
                          WarnAndResubmitExecutor(self._process_request_data)
                        result = process_executor(data)
                    else:
                        result = self._process_request_data(data)
                else:
                    result = data
            except HTTPError as err:
                result = self.request.get_response(err)
            except Exception as err: # catch Exception pylint: disable=W0703
                result = self._handle_unknown_exception(str(err),
                                                        get_traceback())
        return result

    def _get_request_representer(self):
        """
        Returns a representer for the content type specified in the request.

        :raises HTTPUnsupportedMediaType: If the specified content type is
          not supported.
        """
        try:
            mime_type = \
              get_registered_mime_type_for_string(self.request.content_type)
        except KeyError:
            # The client sent a content type we do not support (415).
            raise HTTPUnsupportedMediaType()
        return as_representer(self.context, mime_type)

    def _extract_request_data(self):
        """
        Extracts the data from the representation submitted in the request
        body and returns it.

        This default implementation uses a representer for the content type
        specified by the request to perform the extraction and returns an
        object implementing the
        :class:`everest.representers.interfaces.IResourceDataElement`
        interface.

        :raises HTTPError: To indicate problems with the request data
          extraction in terms of HTTP codes.
        """
        rpr = self._get_request_representer()
        return rpr.data_from_bytes(self.request.body)

    def _process_request_data(self, data):
        """
        Processes the data extracted from the representation.

        Implementations of this method need to check for a conflict caused
        by the request data (e.g., if the slug for a new member in a POST
        request is already used) and call the :meth:`_handle_conflict`
        method in case a conflict was detected.

        :param data: data returned by the :meth:`_extract_request_data`
          method.
        :raises HTTPError: To indicate problems with the request data
          processing in terms of HTTP codes.
        :returns: Response object or dictionary (rendering context).
        """
        raise NotImplementedError('Abstract method.')

    def _handle_empty_body(self):
        """
        Handles requests with an empty body.

        Respond with a 400 "Bad Request".
        """
        http_exc = HTTPBadRequest("Request's body is empty!")
        return self.request.get_response(http_exc)

    def _handle_conflict(self, name):
        """
        Handles requests that triggered a conflict.

        Respond with a 409 "Conflict"
        """
        err = HTTPConflict('Member "%s" already exists!' % name).exception
        return self.request.get_response(err)

    def _status(self, wsgi_http_exc_class):
        """
        Convenience method to obtain a status string from the given HTTP
        exception class.
        """
        return '%(code)s %(title)s' % wsgi_http_exc_class.__dict__


class PutOrPatchResourceView(ModifyingResourceView):
    """
    Base class for views that modify a member through PUT and PATCH requests.
    """
    def _process_request_data(self, data):
        initial_name = self.context.__name__
        self.context.update(data)
        current_name = self.context.__name__
        self.request.response.status = self._status(HTTPOk)
        if initial_name != current_name:
            # FIXME: add conflict detection!
            self._update_response_location_header(self.context)
        # We return the (representation of) the updated member to
        # assist the client in doing the right thing (some clients block
        # access to the Response headers so we may not be able to find the
        # new location when HTTP/1.1 301 is returned).
        return self._get_result(self.context)


class WarnAndResubmitUserMessageChecker(UserMessageChecker):
    """
    Custom user message checker for showing warning messages to the user
    with the possibility for resubmission with warnings suppressed.
    """
    __guid_pattern = re.compile(r".*ignore-message=([a-z0-9\-]{36})")

    def check(self):
        """
        Implements user message checking for views.

        Checks if the current request has an explicit "ignore-message"
        parameter (a GUID) pointing to a message with identical text from a
        previous request, in which case further processing is allowed.
        """
        request = get_current_request()
        ignore_guid = request.params.get('ignore-message')
        coll = request.root['_messages']
        vote = False
        if ignore_guid:
            ignore_mb = coll.get(ignore_guid)
            if not ignore_mb is None and ignore_mb.text == self.message.text:
                vote = True
        return vote

    def create_307_response(self):
        """
        Creates a 307 "Temporary Redirect" response including a HTTP Warning
        header with code 299 that contains the user message received during
        processing the request.
        """
        request = get_current_request()
        msg_mb = UserMessageMember(self.message)
        coll = request.root['_messages']
        coll.add(msg_mb)
        # Figure out the new location URL.
        qs = self.__get_new_query_string(request.query_string,
                                         self.message.slug)
        resubmit_url = "%s?%s" % (request.path_url, qs)
        headers = [('Warning', '299 %s' % self.message.text),
#                       ('Content-Type', cnt_type),
                   ]
        http_exc = HttpWarningResubmit(location=resubmit_url,
                                       detail=self.message.text,
                                       headers=headers)
        return request.get_response(http_exc)

    def __get_new_query_string(self, old_query_string, new_guid):
        # In absence of a function to manipulate the URL query string in place
        # (the request URL params are read-only), resorting to explicit
        # string manipulation seems the safest thing to do.
        if old_query_string:
            # Make sure to *replace* any given ignore message.
            match = self.__guid_pattern.match(old_query_string)
            if match:
                cur_guid = match.groups()[0]
                qs = re.sub(cur_guid, new_guid, old_query_string)
            else:
                qs = "%s&ignore-message=%s" % (old_query_string,
                                               new_guid)
        else:
            qs = "ignore-message=%s" % new_guid
        return qs



"""
Relationships between entities or resources.

This file is part of the everest project.
See LICENSE.txt for licensing, CONTRIBUTORS.txt for contributor information.

Created on Sep 30, 2011.
"""
from everest.constants import CARDINALITY_CONSTANTS
from everest.constants import RELATIONSHIP_DIRECTIONS
from everest.interfaces import IRelationship
from everest.querying.utils import get_filter_specification_factory
from everest.utils import get_nested_attribute
from zope.interface import implementer # pylint: disable=E0611,F0401


__docformat__ = 'reStructuredText en'
__all__ = ['Relationship',
           ]


@implementer(IRelationship)
class Relationship(object):
    """
    Abstract base class for resource and domain relationships.

    A relationship has a source ("relator") and a target ("relatee") and
    implements the relationship operations "ADD", "REMOVE" and "UPDATE".
    """
    def __init__(self, relator, descriptor,
                 direction=RELATIONSHIP_DIRECTIONS.BIDIRECTIONAL):
        self.relator = relator
        self.descriptor = descriptor
        self.direction = direction

    @property
    def specification(self):
        return self.__make_specification()

    def add(self, related, direction=None, safe=False):
        raise NotImplementedError('Abstract method.')

    def remove(self, related, direction=None, safe=False):
        raise NotImplementedError('Abstract method.')

    def _get_specification_attributes(self):
        raise NotImplementedError('Abstract method.')

    def __make_specification(self):
        ref_attr, backref_attr = self._get_specification_attributes()
        #: Builds the filter specification.
        spec_fac = get_filter_specification_factory()
        crd = self.descriptor.cardinality
        if backref_attr is None:
            relatee = get_nested_attribute(self.relator, ref_attr)
            if crd.relatee == CARDINALITY_CONSTANTS.MANY:
                # This is potentially expensive as we may need to iterate over
                # a large collection to create the "contained" specification.
                if len(relatee) > 0:
                    ids = [related.id for related in relatee]
                    spec = spec_fac.create_contained('id', ids)
                else:
                    # Create impossible search criterion.
                    spec = spec_fac.create_equal_to('id', -1)
            else:
                if not relatee is None:
                    spec = spec_fac.create_equal_to('id', relatee.id)
                else:
                    # Create impossible search criterion.
                    spec = spec_fac.create_equal_to('id', -1)
        else:
            if crd.relator == CARDINALITY_CONSTANTS.MANY:
                spec = spec_fac.create_contains(backref_attr, self.relator)
            else:
                spec = spec_fac.create_equal_to(backref_attr, self.relator)
        return spec

    def __str__(self):
        rel_char = '-'
        if self.direction & RELATIONSHIP_DIRECTIONS.FORWARD:
            rel_char += '>'
        if self.direction & RELATIONSHIP_DIRECTIONS.REVERSE:
            rel_char = '<' + rel_char
        return "%s %s%s%s" % (self.__class__.__name__, self.relator,
                              rel_char, self.descriptor.resource_attr)
